# -*- coding: utf-8 -*-
"""
    :codeauthor: Erik Johnson <erik@saltstack.com>
"""

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import platform
import socket
import textwrap

import salt.grains.core as core
import salt.modules.cmdmod
import salt.modules.network
import salt.modules.smbios

# Import Salt Libs
import salt.utils.dns
import salt.utils.files
import salt.utils.network
import salt.utils.path
import salt.utils.platform
from salt._compat import ipaddress

# Import 3rd-party libs
from salt.ext import six
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, mock_open, patch
from tests.support.unit import TestCase, skipIf

# Import Salt Testing Libs
try:
    import pytest
except ImportError as import_error:
    pytest = None


log = logging.getLogger(__name__)

# Globals
IPv4Address = ipaddress.IPv4Address
IPv6Address = ipaddress.IPv6Address
IP4_LOCAL = "127.0.0.1"
IP4_ADD1 = "10.0.0.1"
IP4_ADD2 = "10.0.0.2"
IP6_LOCAL = "::1"
IP6_ADD1 = "2001:4860:4860::8844"
IP6_ADD2 = "2001:4860:4860::8888"
IP6_ADD_SCOPE = "fe80::6238:e0ff:fe06:3f6b%enp2s0"
OS_RELEASE_DIR = os.path.join(os.path.dirname(__file__), "os-releases")
SOLARIS_DIR = os.path.join(os.path.dirname(__file__), "solaris")


@skipIf(not pytest, False)
class CoreGrainsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for core grains
    """

    def setup_loader_modules(self):
        return {core: {}}

    @patch("os.path.isfile")
    def test_parse_etc_os_release(self, path_isfile_mock):
        path_isfile_mock.side_effect = lambda x: x == "/usr/lib/os-release"
        with salt.utils.files.fopen(
            os.path.join(OS_RELEASE_DIR, "ubuntu-17.10")
        ) as os_release_file:
            os_release_content = os_release_file.read()
        with patch("salt.utils.files.fopen", mock_open(read_data=os_release_content)):
            os_release = core._parse_os_release(
                "/etc/os-release", "/usr/lib/os-release"
            )
        self.assertEqual(
            os_release,
            {
                "NAME": "Ubuntu",
                "VERSION": "17.10 (Artful Aardvark)",
                "ID": "ubuntu",
                "ID_LIKE": "debian",
                "PRETTY_NAME": "Ubuntu 17.10",
                "VERSION_ID": "17.10",
                "HOME_URL": "https://www.ubuntu.com/",
                "SUPPORT_URL": "https://help.ubuntu.com/",
                "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
                "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",
                "VERSION_CODENAME": "artful",
                "UBUNTU_CODENAME": "artful",
            },
        )

    def test_parse_cpe_name_wfn(self):
        """
        Parse correct CPE_NAME data WFN formatted
        :return:
        """
        for cpe, cpe_ret in [
            (
                "cpe:/o:opensuse:leap:15.0",
                {
                    "phase": None,
                    "version": "15.0",
                    "product": "leap",
                    "vendor": "opensuse",
                    "part": "operating system",
                },
            ),
            (
                "cpe:/o:vendor:product:42:beta",
                {
                    "phase": "beta",
                    "version": "42",
                    "product": "product",
                    "vendor": "vendor",
                    "part": "operating system",
                },
            ),
        ]:
            ret = core._parse_cpe_name(cpe)
            for key in cpe_ret:
                assert key in ret
                assert cpe_ret[key] == ret[key]

    def test_parse_cpe_name_v23(self):
        """
        Parse correct CPE_NAME data v2.3 formatted
        :return:
        """
        for cpe, cpe_ret in [
            (
                "cpe:2.3:o:microsoft:windows_xp:5.1.601:beta:*:*:*:*:*:*",
                {
                    "phase": "beta",
                    "version": "5.1.601",
                    "product": "windows_xp",
                    "vendor": "microsoft",
                    "part": "operating system",
                },
            ),
            (
                "cpe:2.3:h:corellian:millenium_falcon:1.0:*:*:*:*:*:*:*",
                {
                    "phase": None,
                    "version": "1.0",
                    "product": "millenium_falcon",
                    "vendor": "corellian",
                    "part": "hardware",
                },
            ),
            (
                "cpe:2.3:*:dark_empire:light_saber:3.0:beta:*:*:*:*:*:*",
                {
                    "phase": "beta",
                    "version": "3.0",
                    "product": "light_saber",
                    "vendor": "dark_empire",
                    "part": None,
                },
            ),
        ]:
            ret = core._parse_cpe_name(cpe)
            for key in cpe_ret:
                assert key in ret
                assert cpe_ret[key] == ret[key]

    def test_parse_cpe_name_broken(self):
        """
        Parse broken CPE_NAME data
        :return:
        """
        for cpe in [
            "cpe:broken",
            "cpe:broken:in:all:ways:*:*:*:*",
            "cpe:x:still:broken:123",
            "who:/knows:what:is:here",
        ]:
            assert core._parse_cpe_name(cpe) == {}

    def test_missing_os_release(self):
        with patch("salt.utils.files.fopen", mock_open(read_data={})):
            os_release = core._parse_os_release(
                "/etc/os-release", "/usr/lib/os-release"
            )
        self.assertEqual(os_release, {})

    @skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
    def test__windows_platform_data(self):
        grains = core._windows_platform_data()
        keys = [
            "biosversion",
            "osrelease",
            "kernelrelease",
            "motherboard",
            "serialnumber",
            "timezone",
            "manufacturer",
            "kernelversion",
            "osservicepack",
            "virtual",
            "productname",
            "osfullname",
            "osmanufacturer",
            "osversion",
            "windowsdomain",
        ]
        for key in keys:
            self.assertIn(key, grains)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_gnu_slash_linux_in_os_name(self):
        """
        Test to return a list of all enabled services
        """
        _path_exists_map = {"/proc/1/cmdline": False}
        _path_isfile_map = {}
        _cmd_run_map = {
            "dpkg --print-architecture": "amd64",
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        path_isfile_mock = MagicMock(
            side_effect=lambda x: _path_isfile_map.get(x, False)
        )
        cmd_run_mock = MagicMock(side_effect=lambda x: _cmd_run_map[x])
        empty_mock = MagicMock(return_value={})

        orig_import = __import__
        if six.PY2:
            built_in = "__builtin__"
        else:
            built_in = "builtins"

        def _import_mock(name, *args):
            if name == "lsb_release":
                raise ImportError("No module named lsb_release")
            return orig_import(name, *args)

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Make a bunch of functions return empty dicts, we don't care about
        #   these grains for the purposes of this test.
        # - Mock the osarch
        distro_mock = MagicMock(return_value=("Debian GNU/Linux", "8.3", ""))
        with patch.object(
            salt.utils.platform, "is_proxy", MagicMock(return_value=False)
        ), patch.object(
            core, "_linux_bin_exists", MagicMock(return_value=False)
        ), patch.object(
            os.path, "exists", path_exists_mock
        ), patch(
            "{0}.__import__".format(built_in), side_effect=_import_mock
        ), patch.object(
            os.path, "isfile", path_isfile_mock
        ), patch.object(
            core, "_parse_lsb_release", empty_mock
        ), patch.object(
            core, "_parse_os_release", empty_mock
        ), patch.object(
            core, "_parse_lsb_release", empty_mock
        ), patch.object(
            core, "linux_distribution", distro_mock
        ), patch.object(
            core, "_linux_cpudata", empty_mock
        ), patch.object(
            core, "_linux_gpu_data", empty_mock
        ), patch.object(
            core, "_memdata", empty_mock
        ), patch.object(
            core, "_hw_data", empty_mock
        ), patch.object(
            core, "_virtual", empty_mock
        ), patch.object(
            core, "_ps", empty_mock
        ), patch.dict(
            core.__salt__, {"cmd.run": cmd_run_mock}
        ):
            os_grains = core.os_data()

        self.assertEqual(os_grains.get("os_family"), "Debian")

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_from_cpe_data(self):
        """
        Test if 'os' grain is parsed from CPE_NAME of /etc/os-release
        """
        _path_exists_map = {"/proc/1/cmdline": False}
        _os_release_map = {
            "NAME": "SLES",
            "VERSION": "12-SP1",
            "VERSION_ID": "12.1",
            "PRETTY_NAME": "SUSE Linux Enterprise Server 12 SP1",
            "ID": "sles",
            "ANSI_COLOR": "0;32",
            "CPE_NAME": "cpe:/o:suse:sles:12:sp1",
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        empty_mock = MagicMock(return_value={})
        osarch_mock = MagicMock(return_value="amd64")
        os_release_mock = MagicMock(return_value=_os_release_map)

        orig_import = __import__
        if six.PY2:
            built_in = "__builtin__"
        else:
            built_in = "builtins"

        def _import_mock(name, *args):
            if name == "lsb_release":
                raise ImportError("No module named lsb_release")
            return orig_import(name, *args)

        distro_mock = MagicMock(
            return_value=("SUSE Linux Enterprise Server ", "12", "x86_64")
        )

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Mock the osarch
        with patch.object(
            salt.utils.platform, "is_proxy", MagicMock(return_value=False)
        ), patch.object(
            core, "_linux_bin_exists", MagicMock(return_value=False)
        ), patch.object(
            os.path, "exists", path_exists_mock
        ), patch(
            "{0}.__import__".format(built_in), side_effect=_import_mock
        ), patch.object(
            os.path, "isfile", MagicMock(return_value=False)
        ), patch.object(
            core, "_parse_os_release", os_release_mock
        ), patch.object(
            core, "_parse_lsb_release", empty_mock
        ), patch.object(
            core, "linux_distribution", distro_mock
        ), patch.object(
            core, "_linux_gpu_data", empty_mock
        ), patch.object(
            core, "_hw_data", empty_mock
        ), patch.object(
            core, "_linux_cpudata", empty_mock
        ), patch.object(
            core, "_virtual", empty_mock
        ), patch.dict(
            core.__salt__, {"cmd.run": osarch_mock}
        ):
            os_grains = core.os_data()

        self.assertEqual(os_grains.get("os_family"), "Suse")
        self.assertEqual(os_grains.get("os"), "SUSE")

    def _run_os_grains_tests(self, os_release_filename, os_release_map, expectation):
        path_isfile_mock = MagicMock(
            side_effect=lambda x: x in os_release_map.get("files", [])
        )
        empty_mock = MagicMock(return_value={})
        osarch_mock = MagicMock(return_value="amd64")
        if os_release_filename:
            os_release_data = core._parse_os_release(
                os.path.join(OS_RELEASE_DIR, os_release_filename)
            )
        else:
            os_release_data = os_release_map.get("os_release_file", {})
        os_release_mock = MagicMock(return_value=os_release_data)

        orig_import = __import__
        if six.PY2:
            built_in = "__builtin__"
        else:
            built_in = "builtins"

        def _import_mock(name, *args):
            if name == "lsb_release":
                raise ImportError("No module named lsb_release")
            return orig_import(name, *args)

        suse_release_file = os_release_map.get("suse_release_file")

        file_contents = {"/proc/1/cmdline": ""}
        if suse_release_file:
            file_contents["/etc/SuSE-release"] = suse_release_file

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Mock the osarch
        distro_mock = MagicMock(return_value=os_release_map["linux_distribution"])
        with patch.object(
            salt.utils.platform, "is_proxy", MagicMock(return_value=False)
        ), patch.object(
            core, "_linux_bin_exists", MagicMock(return_value=False)
        ), patch.object(
            os.path, "exists", path_isfile_mock
        ), patch(
            "{0}.__import__".format(built_in), side_effect=_import_mock
        ), patch.object(
            os.path, "isfile", path_isfile_mock
        ), patch.object(
            core, "_parse_os_release", os_release_mock
        ), patch.object(
            core, "_parse_lsb_release", empty_mock
        ), patch(
            "salt.utils.files.fopen", mock_open(read_data=file_contents)
        ), patch.object(
            core, "linux_distribution", distro_mock
        ), patch.object(
            core, "_linux_gpu_data", empty_mock
        ), patch.object(
            core, "_linux_cpudata", empty_mock
        ), patch.object(
            core, "_virtual", empty_mock
        ), patch.dict(
            core.__salt__, {"cmd.run": osarch_mock}
        ):
            os_grains = core.os_data()

        grains = {
            k: v
            for k, v in os_grains.items()
            if k
            in set(
                [
                    "os",
                    "os_family",
                    "osfullname",
                    "oscodename",
                    "osfinger",
                    "osrelease",
                    "osrelease_info",
                    "osmajorrelease",
                ]
            )
        }
        self.assertEqual(grains, expectation)

    def _run_suse_os_grains_tests(self, os_release_map, expectation):
        os_release_map["linux_distribution"] = ("SUSE test", "version", "arch")
        expectation["os"] = "SUSE"
        expectation["os_family"] = "Suse"
        self._run_os_grains_tests(None, os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_sles11sp3(self):
        """
        Test if OS grains are parsed correctly in SLES 11 SP3
        """
        _os_release_map = {
            "suse_release_file": textwrap.dedent(
                """
                SUSE Linux Enterprise Server 11 (x86_64)
                VERSION = 11
                PATCHLEVEL = 3
                """
            ),
            "files": ["/etc/SuSE-release"],
        }
        expectation = {
            "oscodename": "SUSE Linux Enterprise Server 11 SP3",
            "osfullname": "SLES",
            "osrelease": "11.3",
            "osrelease_info": (11, 3),
            "osmajorrelease": 11,
            "osfinger": "SLES-11",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_sles11sp4(self):
        """
        Test if OS grains are parsed correctly in SLES 11 SP4
        """
        _os_release_map = {
            "os_release_file": {
                "NAME": "SLES",
                "VERSION": "11.4",
                "VERSION_ID": "11.4",
                "PRETTY_NAME": "SUSE Linux Enterprise Server 11 SP4",
                "ID": "sles",
                "ANSI_COLOR": "0;32",
                "CPE_NAME": "cpe:/o:suse:sles:11:4",
            },
        }
        expectation = {
            "oscodename": "SUSE Linux Enterprise Server 11 SP4",
            "osfullname": "SLES",
            "osrelease": "11.4",
            "osrelease_info": (11, 4),
            "osmajorrelease": 11,
            "osfinger": "SLES-11",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_sles12(self):
        """
        Test if OS grains are parsed correctly in SLES 12
        """
        _os_release_map = {
            "os_release_file": {
                "NAME": "SLES",
                "VERSION": "12",
                "VERSION_ID": "12",
                "PRETTY_NAME": "SUSE Linux Enterprise Server 12",
                "ID": "sles",
                "ANSI_COLOR": "0;32",
                "CPE_NAME": "cpe:/o:suse:sles:12",
            },
        }
        expectation = {
            "oscodename": "SUSE Linux Enterprise Server 12",
            "osfullname": "SLES",
            "osrelease": "12",
            "osrelease_info": (12,),
            "osmajorrelease": 12,
            "osfinger": "SLES-12",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_sles12sp1(self):
        """
        Test if OS grains are parsed correctly in SLES 12 SP1
        """
        _os_release_map = {
            "os_release_file": {
                "NAME": "SLES",
                "VERSION": "12-SP1",
                "VERSION_ID": "12.1",
                "PRETTY_NAME": "SUSE Linux Enterprise Server 12 SP1",
                "ID": "sles",
                "ANSI_COLOR": "0;32",
                "CPE_NAME": "cpe:/o:suse:sles:12:sp1",
            },
        }
        expectation = {
            "oscodename": "SUSE Linux Enterprise Server 12 SP1",
            "osfullname": "SLES",
            "osrelease": "12.1",
            "osrelease_info": (12, 1),
            "osmajorrelease": 12,
            "osfinger": "SLES-12",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_opensuse_leap_42_1(self):
        """
        Test if OS grains are parsed correctly in openSUSE Leap 42.1
        """
        _os_release_map = {
            "os_release_file": {
                "NAME": "openSUSE Leap",
                "VERSION": "42.1",
                "VERSION_ID": "42.1",
                "PRETTY_NAME": "openSUSE Leap 42.1 (x86_64)",
                "ID": "opensuse",
                "ANSI_COLOR": "0;32",
                "CPE_NAME": "cpe:/o:opensuse:opensuse:42.1",
            },
        }
        expectation = {
            "oscodename": "openSUSE Leap 42.1 (x86_64)",
            "osfullname": "Leap",
            "osrelease": "42.1",
            "osrelease_info": (42, 1),
            "osmajorrelease": 42,
            "osfinger": "Leap-42",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_suse_os_grains_tumbleweed(self):
        """
        Test if OS grains are parsed correctly in openSUSE Tumbleweed
        """
        _os_release_map = {
            "os_release_file": {
                "NAME": "openSUSE",
                "VERSION": "Tumbleweed",
                "VERSION_ID": "20160504",
                "PRETTY_NAME": "openSUSE Tumbleweed (20160504) (x86_64)",
                "ID": "opensuse",
                "ANSI_COLOR": "0;32",
                "CPE_NAME": "cpe:/o:opensuse:opensuse:20160504",
            },
        }
        expectation = {
            "oscodename": "openSUSE Tumbleweed (20160504) (x86_64)",
            "osfullname": "Tumbleweed",
            "osrelease": "20160504",
            "osrelease_info": (20160504,),
            "osmajorrelease": 20160504,
            "osfinger": "Tumbleweed-20160504",
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_debian_7_os_grains(self):
        """
        Test if OS grains are parsed correctly in Debian 7 "wheezy"
        """
        _os_release_map = {
            "linux_distribution": ("debian", "7.11", ""),
        }
        expectation = {
            "os": "Debian",
            "os_family": "Debian",
            "oscodename": "wheezy",
            "osfullname": "Debian GNU/Linux",
            "osrelease": "7",
            "osrelease_info": (7,),
            "osmajorrelease": 7,
            "osfinger": "Debian-7",
        }
        self._run_os_grains_tests("debian-7", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_debian_8_os_grains(self):
        """
        Test if OS grains are parsed correctly in Debian 8 "jessie"
        """
        _os_release_map = {
            "linux_distribution": ("debian", "8.10", ""),
        }
        expectation = {
            "os": "Debian",
            "os_family": "Debian",
            "oscodename": "jessie",
            "osfullname": "Debian GNU/Linux",
            "osrelease": "8",
            "osrelease_info": (8,),
            "osmajorrelease": 8,
            "osfinger": "Debian-8",
        }
        self._run_os_grains_tests("debian-8", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_debian_9_os_grains(self):
        """
        Test if OS grains are parsed correctly in Debian 9 "stretch"
        """
        _os_release_map = {
            "linux_distribution": ("debian", "9.3", ""),
        }
        expectation = {
            "os": "Debian",
            "os_family": "Debian",
            "oscodename": "stretch",
            "osfullname": "Debian GNU/Linux",
            "osrelease": "9",
            "osrelease_info": (9,),
            "osmajorrelease": 9,
            "osfinger": "Debian-9",
        }
        self._run_os_grains_tests("debian-9", _os_release_map, expectation)

    def test_unicode_error(self):
        raise_unicode_mock = MagicMock(
            name="raise_unicode_error", side_effect=UnicodeError
        )
        with patch("salt.grains.core.hostname"):
            with patch("socket.getaddrinfo", raise_unicode_mock):
                ret = salt.grains.core.ip_fqdn()
                assert ret["fqdn_ip4"] == ret["fqdn_ip6"] == []

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_ubuntu_xenial_os_grains(self):
        """
        Test if OS grains are parsed correctly in Ubuntu 16.04 "Xenial Xerus"
        """
        _os_release_map = {
            "linux_distribution": ("Ubuntu", "16.04", "xenial"),
        }
        expectation = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "xenial",
            "osfullname": "Ubuntu",
            "osrelease": "16.04",
            "osrelease_info": (16, 4),
            "osmajorrelease": 16,
            "osfinger": "Ubuntu-16.04",
        }
        self._run_os_grains_tests("ubuntu-16.04", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_ubuntu_artful_os_grains(self):
        """
        Test if OS grains are parsed correctly in Ubuntu 17.10 "Artful Aardvark"
        """
        _os_release_map = {
            "linux_distribution": ("Ubuntu", "17.10", "artful"),
        }
        expectation = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "artful",
            "osfullname": "Ubuntu",
            "osrelease": "17.10",
            "osrelease_info": (17, 10),
            "osmajorrelease": 17,
            "osfinger": "Ubuntu-17.10",
        }
        self._run_os_grains_tests("ubuntu-17.10", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
    def test_windows_platform_data(self):
        """
        Test the _windows_platform_data function
        """
        grains = [
            "biosversion",
            "kernelrelease",
            "kernelversion",
            "manufacturer",
            "motherboard",
            "osfullname",
            "osmanufacturer",
            "osrelease",
            "osservicepack",
            "osversion",
            "productname",
            "serialnumber",
            "timezone",
            "virtual",
            "windowsdomain",
            "windowsdomaintype",
        ]
        returned_grains = core._windows_platform_data()
        for grain in grains:
            self.assertIn(grain, returned_grains)

        valid_types = ["Unknown", "Unjoined", "Workgroup", "Domain"]
        self.assertIn(returned_grains["windowsdomaintype"], valid_types)
        valid_releases = [
            "Vista",
            "7",
            "8",
            "8.1",
            "10",
            "2008Server",
            "2008ServerR2",
            "2012Server",
            "2012ServerR2",
            "2016Server",
            "2019Server",
        ]
        self.assertIn(returned_grains["osrelease"], valid_releases)

    def test__windows_os_release_grain(self):
        versions = {
            "Windows 10 Home": "10",
            "Windows 10 Pro": "10",
            "Windows 10 Pro for Workstations": "10",
            "Windows 10 Pro Education": "10",
            "Windows 10 Enterprise": "10",
            "Windows 10 Enterprise LTSB": "10",
            "Windows 10 Education": "10",
            "Windows 10 IoT Core": "10",
            "Windows 10 IoT Enterprise": "10",
            "Windows 10 S": "10",
            "Windows 8.1": "8.1",
            "Windows 8.1 Pro": "8.1",
            "Windows 8.1 Enterprise": "8.1",
            "Windows 8.1 OEM": "8.1",
            "Windows 8.1 with Bing": "8.1",
            "Windows 8": "8",
            "Windows 8 Pro": "8",
            "Windows 8 Enterprise": "8",
            "Windows 8 OEM": "8",
            "Windows 7 Starter": "7",
            "Windows 7 Home Basic": "7",
            "Windows 7 Home Premium": "7",
            "Windows 7 Professional": "7",
            "Windows 7 Enterprise": "7",
            "Windows 7 Ultimate": "7",
            "Windows Thin PC": "Thin",
            "Windows Vista Starter": "Vista",
            "Windows Vista Home Basic": "Vista",
            "Windows Vista Home Premium": "Vista",
            "Windows Vista Business": "Vista",
            "Windows Vista Enterprise": "Vista",
            "Windows Vista Ultimate": "Vista",
            "Windows Server 2019 Essentials": "2019Server",
            "Windows Server 2019 Standard": "2019Server",
            "Windows Server 2019 Datacenter": "2019Server",
            "Windows Server 2016 Essentials": "2016Server",
            "Windows Server 2016 Standard": "2016Server",
            "Windows Server 2016 Datacenter": "2016Server",
            "Windows Server 2012 R2 Foundation": "2012ServerR2",
            "Windows Server 2012 R2 Essentials": "2012ServerR2",
            "Windows Server 2012 R2 Standard": "2012ServerR2",
            "Windows Server 2012 R2 Datacenter": "2012ServerR2",
            "Windows Server 2012 Foundation": "2012Server",
            "Windows Server 2012 Essentials": "2012Server",
            "Windows Server 2012 Standard": "2012Server",
            "Windows Server 2012 Datacenter": "2012Server",
            "Windows MultiPoint Server 2012": "2012Server",
            "Windows Small Business Server 2011": "2011Server",
            "Windows MultiPoint Server 2011": "2011Server",
            "Windows Home Server 2011": "2011Server",
            "Windows MultiPoint Server 2010": "2010Server",
            "Windows Server 2008 R2 Foundation": "2008ServerR2",
            "Windows Server 2008 R2 Standard": "2008ServerR2",
            "Windows Server 2008 R2 Enterprise": "2008ServerR2",
            "Windows Server 2008 R2 Datacenter": "2008ServerR2",
            "Windows Server 2008 R2 for Itanium-based Systems": "2008ServerR2",
            "Windows Web Server 2008 R2": "2008ServerR2",
            "Windows Storage Server 2008 R2": "2008ServerR2",
            "Windows HPC Server 2008 R2": "2008ServerR2",
            "Windows Server 2008 Standard": "2008Server",
            "Windows Server 2008 Enterprise": "2008Server",
            "Windows Server 2008 Datacenter": "2008Server",
            "Windows Server 2008 for Itanium-based Systems": "2008Server",
            "Windows Server Foundation 2008": "2008Server",
            "Windows Essential Business Server 2008": "2008Server",
            "Windows HPC Server 2008": "2008Server",
            "Windows Small Business Server 2008": "2008Server",
            "Windows Storage Server 2008": "2008Server",
            "Windows Web Server 2008": "2008Server",
        }
        for caption in versions:
            version = core._windows_os_release_grain(caption, 1)
            self.assertEqual(
                version,
                versions[caption],
                "version: {0}\n"
                "found: {1}\n"
                "caption: {2}".format(version, versions[caption], caption),
            )

        embedded_versions = {
            "Windows Embedded 8.1 Industry Pro": "8.1",
            "Windows Embedded 8 Industry Pro": "8",
            "Windows POSReady 7": "7",
            "Windows Embedded Standard 7": "7",
            "Windows Embedded POSReady 2009": "2009",
            "Windows Embedded Standard 2009": "2009",
            "Windows XP Embedded": "XP",
        }
        for caption in embedded_versions:
            version = core._windows_os_release_grain(caption, 1)
            self.assertEqual(
                version,
                embedded_versions[caption],
                "{0} != {1}\n"
                "version: {0}\n"
                "found: {1}\n"
                "caption: {2}".format(version, embedded_versions[caption], caption),
            )

        # Special Cases
        # Windows Embedded Standard is Windows 7
        caption = "Windows Embedded Standard"
        with patch("platform.release", MagicMock(return_value="7")):
            version = core._windows_os_release_grain(caption, 1)
            self.assertEqual(version, "7")

        # Microsoft Hyper-V Server 2019
        # Issue https://github.com/saltstack/salt/issue/55212
        caption = "Microsoft Hyper-V Server"
        version = core._windows_os_release_grain(caption, 1)
        self.assertEqual(version, "2019Server")

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_linux_memdata(self):
        """
        Test memdata on Linux systems
        """
        _proc_meminfo = textwrap.dedent(
            """\
            MemTotal:       16277028 kB
            SwapTotal:       4789244 kB"""
        )
        with patch("salt.utils.files.fopen", mock_open(read_data=_proc_meminfo)):
            memdata = core._linux_memdata()
        self.assertEqual(memdata.get("mem_total"), 15895)
        self.assertEqual(memdata.get("swap_total"), 4676)

    @skipIf(salt.utils.platform.is_windows(), "System is Windows")
    def test_bsd_memdata(self):
        """
        Test to memdata on *BSD systems
        """
        _path_exists_map = {}
        _path_isfile_map = {}
        _cmd_run_map = {
            "freebsd-version -u": "10.3-RELEASE",
            "/sbin/sysctl -n hw.physmem": "2121781248",
            "/sbin/sysctl -n vm.swap_total": "419430400",
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        path_isfile_mock = MagicMock(
            side_effect=lambda x: _path_isfile_map.get(x, False)
        )
        cmd_run_mock = MagicMock(side_effect=lambda x: _cmd_run_map[x])
        empty_mock = MagicMock(return_value={})

        mock_freebsd_uname = (
            "FreeBSD",
            "freebsd10.3-hostname-8148",
            "10.3-RELEASE",
            "FreeBSD 10.3-RELEASE #0 r297264: Fri Mar 25 02:10:02 UTC 2016     root@releng1.nyi.freebsd.org:/usr/obj/usr/src/sys/GENERIC",
            "amd64",
            "amd64",
        )

        with patch("platform.uname", MagicMock(return_value=mock_freebsd_uname)):
            with patch.object(
                salt.utils.platform, "is_linux", MagicMock(return_value=False)
            ):
                with patch.object(
                    salt.utils.platform, "is_freebsd", MagicMock(return_value=True)
                ):
                    # Skip the first if statement
                    with patch.object(
                        salt.utils.platform, "is_proxy", MagicMock(return_value=False)
                    ):
                        # Skip the init grain compilation (not pertinent)
                        with patch.object(os.path, "exists", path_exists_mock):
                            with patch("salt.utils.path.which") as mock:
                                mock.return_value = "/sbin/sysctl"
                                # Make a bunch of functions return empty dicts,
                                # we don't care about these grains for the
                                # purposes of this test.
                                with patch.object(core, "_bsd_cpudata", empty_mock):
                                    with patch.object(core, "_hw_data", empty_mock):
                                        with patch.object(core, "_virtual", empty_mock):
                                            with patch.object(core, "_ps", empty_mock):
                                                # Mock the osarch
                                                with patch.dict(
                                                    core.__salt__,
                                                    {"cmd.run": cmd_run_mock},
                                                ):
                                                    os_grains = core.os_data()

        self.assertEqual(os_grains.get("mem_total"), 2023)
        self.assertEqual(os_grains.get("swap_total"), 400)

    @skipIf(salt.utils.platform.is_windows(), "System is Windows")
    def test_docker_virtual(self):
        """
        Test if virtual grains are parsed correctly in Docker.
        """
        with patch.object(os.path, "isdir", MagicMock(return_value=False)):
            with patch.object(
                os.path,
                "isfile",
                MagicMock(
                    side_effect=lambda x: True if x == "/proc/1/cgroup" else False
                ),
            ):
                for cgroup_substr in (
                    ":/system.slice/docker",
                    ":/docker/",
                    ":/docker-ce/",
                ):
                    cgroup_data = "10:memory{0}a_long_sha256sum".format(cgroup_substr)
                    log.debug("Testing Docker cgroup substring '%s'", cgroup_substr)
                    with patch(
                        "salt.utils.files.fopen", mock_open(read_data=cgroup_data)
                    ):
                        with patch.dict(core.__salt__, {"cmd.run_all": MagicMock()}):
                            grains = core._virtual({"kernel": "Linux"})
                            self.assertEqual(grains.get("virtual_subtype"), "Docker")
                            self.assertEqual(
                                grains.get("virtual"), "container",
                            )

    @skipIf(salt.utils.platform.is_windows(), "System is Windows")
    def test_lxc_virtual(self):
        """
        Test if virtual grains are parsed correctly in LXC.
        """
        with patch.object(os.path, "isdir", MagicMock(return_value=False)):
            with patch.object(
                os.path,
                "isfile",
                MagicMock(
                    side_effect=lambda x: True if x == "/proc/1/cgroup" else False
                ),
            ):
                cgroup_data = "10:memory:/lxc/a_long_sha256sum"
                with patch("salt.utils.files.fopen", mock_open(read_data=cgroup_data)):
                    with patch.dict(core.__salt__, {"cmd.run_all": MagicMock()}):
                        grains = core._virtual({"kernel": "Linux"})
                        self.assertEqual(grains.get("virtual_subtype"), "LXC")
                        self.assertEqual(
                            grains.get("virtual"), "container",
                        )

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_xen_virtual(self):
        """
        Test if OS grains are parsed correctly in Ubuntu Xenial Xerus
        """
        with patch.multiple(
            os.path,
            isdir=MagicMock(side_effect=lambda x: x == "/sys/bus/xen"),
            isfile=MagicMock(
                side_effect=lambda x: x == "/sys/bus/xen/drivers/xenconsole"
            ),
        ):
            with patch.dict(core.__salt__, {"cmd.run": MagicMock(return_value="")}):
                log.debug("Testing Xen")
                self.assertEqual(
                    core._virtual({"kernel": "Linux"}).get("virtual_subtype"),
                    "Xen PV DomU",
                )

    def test_if_virtual_subtype_exists_virtual_should_fallback_to_virtual(self):
        def mockstat(path):
            if path == "/":
                return "fnord"
            elif path == "/proc/1/root/.":
                return "roscivs"
            return None

        with patch.dict(
            core.__salt__,
            {
                "cmd.run": MagicMock(return_value=""),
                "cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": ""}),
            },
        ):
            with patch.multiple(
                os.path,
                isfile=MagicMock(return_value=False),
                isdir=MagicMock(side_effect=lambda x: x == "/proc"),
            ):
                with patch.multiple(
                    os, stat=MagicMock(side_effect=mockstat),
                ):
                    grains = core._virtual({"kernel": "Linux"})
                    assert grains.get("virtual_subtype") is not None
                    assert grains.get("virtual") == "virtual"

    def _check_ipaddress(self, value, ip_v):
        """
        check if ip address in a list is valid
        """
        for val in value:
            assert isinstance(val, six.string_types)
            ip_method = "is_ipv{0}".format(ip_v)
            self.assertTrue(getattr(salt.utils.network, ip_method)(val))

    def _check_empty(self, key, value, empty):
        """
        if empty is False and value does not exist assert error
        if empty is True and value exists assert error
        """
        if not empty and not value:
            raise Exception("{0} is empty, expecting a value".format(key))
        elif empty and value:
            raise Exception(
                "{0} is suppose to be empty. value: {1} \
                            exists".format(
                    key, value
                )
            )

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_fqdn_return(self):
        """
        test ip4 and ip6 return values
        """
        net_ip4_mock = [IP4_LOCAL, IP4_ADD1, IP4_ADD2]
        net_ip6_mock = [IP6_LOCAL, IP6_ADD1, IP6_ADD2]

        self._run_fqdn_tests(
            net_ip4_mock, net_ip6_mock, ip4_empty=False, ip6_empty=False
        )

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_fqdn6_empty(self):
        """
        test when ip6 is empty
        """
        net_ip4_mock = [IP4_LOCAL, IP4_ADD1, IP4_ADD2]
        net_ip6_mock = []

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock, ip4_empty=False)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_fqdn4_empty(self):
        """
        test when ip4 is empty
        """
        net_ip4_mock = []
        net_ip6_mock = [IP6_LOCAL, IP6_ADD1, IP6_ADD2]

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock, ip6_empty=False)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_fqdn_all_empty(self):
        """
        test when both ip4 and ip6 are empty
        """
        net_ip4_mock = []
        net_ip6_mock = []

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock)

    def _run_fqdn_tests(
        self, net_ip4_mock, net_ip6_mock, ip6_empty=True, ip4_empty=True
    ):
        def _check_type(key, value, ip4_empty, ip6_empty):
            """
            check type and other checks
            """
            assert isinstance(value, list)

            if "4" in key:
                self._check_empty(key, value, ip4_empty)
                self._check_ipaddress(value, ip_v="4")
            elif "6" in key:
                self._check_empty(key, value, ip6_empty)
                self._check_ipaddress(value, ip_v="6")

        ip4_mock = [(2, 1, 6, "", (IP4_ADD1, 0)), (2, 3, 0, "", (IP4_ADD2, 0))]
        ip6_mock = [
            (10, 1, 6, "", (IP6_ADD1, 0, 0, 0)),
            (10, 3, 0, "", (IP6_ADD2, 0, 0, 0)),
        ]

        with patch.dict(core.__opts__, {"ipv6": False}):
            with patch.object(
                salt.utils.network, "ip_addrs", MagicMock(return_value=net_ip4_mock)
            ):
                with patch.object(
                    salt.utils.network,
                    "ip_addrs6",
                    MagicMock(return_value=net_ip6_mock),
                ):
                    with patch.object(
                        core.socket, "getaddrinfo", side_effect=[ip4_mock, ip6_mock]
                    ):
                        get_fqdn = core.ip_fqdn()
                        ret_keys = ["fqdn_ip4", "fqdn_ip6", "ipv4", "ipv6"]
                        for key in ret_keys:
                            value = get_fqdn[key]
                            _check_type(key, value, ip4_empty, ip6_empty)

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    @patch.object(salt.utils.platform, "is_windows", MagicMock(return_value=False))
    @patch("salt.grains.core.__opts__", {"ipv6": False})
    def test_dns_return(self):
        """
        test the return for a dns grain. test for issue:
        https://github.com/saltstack/salt/issues/41230
        """
        resolv_mock = {
            "domain": "",
            "sortlist": [],
            "nameservers": [
                ipaddress.IPv4Address(IP4_ADD1),
                ipaddress.IPv6Address(IP6_ADD1),
                IP6_ADD_SCOPE,
            ],
            "ip4_nameservers": [ipaddress.IPv4Address(IP4_ADD1)],
            "search": ["test.saltstack.com"],
            "ip6_nameservers": [ipaddress.IPv6Address(IP6_ADD1), IP6_ADD_SCOPE],
            "options": [],
        }
        ret = {
            "dns": {
                "domain": "",
                "sortlist": [],
                "nameservers": [IP4_ADD1, IP6_ADD1, IP6_ADD_SCOPE],
                "ip4_nameservers": [IP4_ADD1],
                "search": ["test.saltstack.com"],
                "ip6_nameservers": [IP6_ADD1, IP6_ADD_SCOPE],
                "options": [],
            }
        }
        with patch.object(
            salt.utils.dns, "parse_resolv", MagicMock(return_value=resolv_mock)
        ):
            assert core.dns() == ret

    def test_enablefqdnsFalse(self):
        """
        tests enable_fqdns_grains is set to False
        """
        with patch.dict("salt.grains.core.__opts__", {"enable_fqdns_grains": False}):
            assert core.fqdns() == {"fqdns": []}

    def test_enablefqdnsTrue(self):
        """
        testing that grains uses network.fqdns module
        """
        with patch.dict(
            "salt.grains.core.__salt__",
            {"network.fqdns": MagicMock(return_value="my.fake.domain")},
        ):
            with patch.dict("salt.grains.core.__opts__", {"enable_fqdns_grains": True}):
                assert core.fqdns() == "my.fake.domain"

    def test_enablefqdnsNone(self):
        """
        testing default fqdns grains is returned when enable_fqdns_grains is None
        """
        with patch.dict("salt.grains.core.__opts__", {"enable_fqdns_grains": None}):
            assert core.fqdns() == {"fqdns": []}

    def test_enablefqdnswithoutpaching(self):
        """
        testing fqdns grains is enabled by default
        """
        with patch.dict(
            "salt.grains.core.__salt__",
            {"network.fqdns": MagicMock(return_value="my.fake.domain")},
        ):
            assert core.fqdns() == "my.fake.domain"

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    @patch(
        "salt.utils.network.ip_addrs", MagicMock(return_value=["1.2.3.4", "5.6.7.8"])
    )
    @patch(
        "salt.utils.network.ip_addrs6",
        MagicMock(return_value=["fe80::a8b2:93ff:fe00:0", "fe80::a8b2:93ff:dead:beef"]),
    )
    @patch(
        "salt.utils.network.socket.getfqdn", MagicMock(side_effect=lambda v: v)
    )  # Just pass-through
    def test_fqdns_return(self):
        """
        test the return for a dns grain. test for issue:
        https://github.com/saltstack/salt/issues/41230
        """
        reverse_resolv_mock = [
            ("foo.bar.baz", [], ["1.2.3.4"]),
            ("rinzler.evil-corp.com", [], ["5.6.7.8"]),
            ("foo.bar.baz", [], ["fe80::a8b2:93ff:fe00:0"]),
            ("bluesniff.foo.bar", [], ["fe80::a8b2:93ff:dead:beef"]),
        ]
        ret = {"fqdns": ["bluesniff.foo.bar", "foo.bar.baz", "rinzler.evil-corp.com"]}
        with patch.dict(core.__salt__, {"network.fqdns": salt.modules.network.fqdns}):
            with patch.object(socket, "gethostbyaddr", side_effect=reverse_resolv_mock):
                fqdns = core.fqdns()
                assert "fqdns" in fqdns
                assert len(fqdns["fqdns"]) == len(ret["fqdns"])
                assert set(fqdns["fqdns"]) == set(ret["fqdns"])

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    @patch("salt.utils.network.ip_addrs", MagicMock(return_value=["1.2.3.4"]))
    @patch("salt.utils.network.ip_addrs6", MagicMock(return_value=[]))
    def test_fqdns_socket_error(self):
        """
        test the behavior on non-critical socket errors of the dns grain
        """

        def _gen_gethostbyaddr(errno):
            def _gethostbyaddr(_):
                herror = socket.herror()
                herror.errno = errno
                raise herror

            return _gethostbyaddr

        for errno in (0, core.HOST_NOT_FOUND, core.NO_DATA):
            mock_log = MagicMock()
            with patch.dict(
                core.__salt__, {"network.fqdns": salt.modules.network.fqdns}
            ):
                with patch.object(
                    socket, "gethostbyaddr", side_effect=_gen_gethostbyaddr(errno)
                ):
                    with patch("salt.modules.network.log", mock_log):
                        self.assertEqual(core.fqdns(), {"fqdns": []})
                        mock_log.debug.assert_called()
                        mock_log.error.assert_not_called()

        mock_log = MagicMock()
        with patch.dict(core.__salt__, {"network.fqdns": salt.modules.network.fqdns}):
            with patch.object(
                socket, "gethostbyaddr", side_effect=_gen_gethostbyaddr(-1)
            ):
                with patch("salt.modules.network.log", mock_log):
                    self.assertEqual(core.fqdns(), {"fqdns": []})
                    mock_log.debug.assert_called_once()
                    mock_log.error.assert_called()

    def test_core_virtual(self):
        """
        test virtual grain with cmd virt-what
        """
        virt = "kvm"
        with patch.object(
            salt.utils.platform, "is_windows", MagicMock(return_value=False)
        ):
            with patch.object(salt.utils.path, "which", MagicMock(return_value=True)):
                with patch.dict(
                    core.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={
                                "pid": 78,
                                "retcode": 0,
                                "stderr": "",
                                "stdout": virt,
                            }
                        )
                    },
                ):
                    osdata = {
                        "kernel": "test",
                    }
                    ret = core._virtual(osdata)
                    self.assertEqual(ret["virtual"], virt)

    def test_solaris_sparc_s7zone(self):
        """
        verify productname grain for s7 zone
        """
        expectation = {
            "productname": "SPARC S7-2",
            "product": "SPARC S7-2",
        }
        with salt.utils.files.fopen(
            os.path.join(SOLARIS_DIR, "prtconf.s7-zone")
        ) as sparc_return_data:
            this_sparc_return_data = "\n".join(sparc_return_data.readlines())
            this_sparc_return_data += "\n"
        self._check_solaris_sparc_productname_grains(
            this_sparc_return_data, expectation
        )

    def test_solaris_sparc_s7(self):
        """
        verify productname grain for s7
        """
        expectation = {
            "productname": "SPARC S7-2",
            "product": "SPARC S7-2",
        }
        with salt.utils.files.fopen(
            os.path.join(SOLARIS_DIR, "prtdiag.s7")
        ) as sparc_return_data:
            this_sparc_return_data = "\n".join(sparc_return_data.readlines())
            this_sparc_return_data += "\n"
        self._check_solaris_sparc_productname_grains(
            this_sparc_return_data, expectation
        )

    def test_solaris_sparc_t5220(self):
        """
        verify productname grain for t5220
        """
        expectation = {
            "productname": "SPARC Enterprise T5220",
            "product": "SPARC Enterprise T5220",
        }
        with salt.utils.files.fopen(
            os.path.join(SOLARIS_DIR, "prtdiag.t5220")
        ) as sparc_return_data:
            this_sparc_return_data = "\n".join(sparc_return_data.readlines())
            this_sparc_return_data += "\n"
        self._check_solaris_sparc_productname_grains(
            this_sparc_return_data, expectation
        )

    def test_solaris_sparc_t5220zone(self):
        """
        verify productname grain for t5220 zone
        """
        expectation = {
            "productname": "SPARC Enterprise T5220",
            "product": "SPARC Enterprise T5220",
        }
        with salt.utils.files.fopen(
            os.path.join(SOLARIS_DIR, "prtconf.t5220-zone")
        ) as sparc_return_data:
            this_sparc_return_data = "\n".join(sparc_return_data.readlines())
            this_sparc_return_data += "\n"
        self._check_solaris_sparc_productname_grains(
            this_sparc_return_data, expectation
        )

    def _check_solaris_sparc_productname_grains(self, prtdata, expectation):
        """
        verify product grains on solaris sparc
        """
        import platform

        path_isfile_mock = MagicMock(side_effect=lambda x: x in ["/etc/release"])
        with salt.utils.files.fopen(
            os.path.join(OS_RELEASE_DIR, "solaris-11.3")
        ) as os_release_file:
            os_release_content = os_release_file.readlines()
        uname_mock = MagicMock(
            return_value=("SunOS", "testsystem", "5.11", "11.3", "sunv4", "sparc")
        )
        with patch.object(platform, "uname", uname_mock), patch.object(
            salt.utils.platform, "is_proxy", MagicMock(return_value=False)
        ), patch.object(
            salt.utils.platform, "is_linux", MagicMock(return_value=False)
        ), patch.object(
            salt.utils.platform, "is_windows", MagicMock(return_value=False)
        ), patch.object(
            salt.utils.platform, "is_smartos", MagicMock(return_value=False)
        ), patch.object(
            salt.utils.path, "which_bin", MagicMock(return_value=None)
        ), patch.object(
            os.path, "isfile", path_isfile_mock
        ), patch(
            "salt.utils.files.fopen", mock_open(read_data=os_release_content)
        ) as os_release_file, patch.object(
            core,
            "_sunos_cpudata",
            MagicMock(
                return_value={
                    "cpuarch": "sparcv9",
                    "num_cpus": "1",
                    "cpu_model": "MOCK_CPU_MODEL",
                    "cpu_flags": [],
                }
            ),
        ), patch.object(
            core, "_memdata", MagicMock(return_value={"mem_total": 16384})
        ), patch.object(
            core, "_virtual", MagicMock(return_value={})
        ), patch.object(
            core, "_ps", MagicMock(return_value={})
        ), patch.object(
            salt.utils.path, "which", MagicMock(return_value=True)
        ), patch.dict(
            core.__salt__, {"cmd.run": MagicMock(return_value=prtdata)}
        ):
            os_grains = core.os_data()
        grains = {
            k: v for k, v in os_grains.items() if k in set(["product", "productname"])
        }
        self.assertEqual(grains, expectation)

    @patch("os.path.isfile")
    @patch("os.path.isdir")
    def test_core_virtual_unicode(self, mock_file, mock_dir):
        """
        test virtual grain with unicode character in product_name file
        """

        def path_side_effect(path):
            if path == "/sys/devices/virtual/dmi/id/product_name":
                return True
            return False

        virt = "kvm"
        mock_file.side_effect = path_side_effect
        mock_dir.side_effect = path_side_effect
        with patch.object(
            salt.utils.platform, "is_windows", MagicMock(return_value=False)
        ):
            with patch.object(salt.utils.path, "which", MagicMock(return_value=True)):
                with patch.dict(
                    core.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={
                                "pid": 78,
                                "retcode": 0,
                                "stderr": "",
                                "stdout": virt,
                            }
                        )
                    },
                ):
                    with patch("salt.utils.files.fopen", mock_open(read_data="嗨")):
                        osdata = {
                            "kernel": "Linux",
                        }
                        ret = core._virtual(osdata)
                        self.assertEqual(ret["virtual"], virt)

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/sbin/sysctl"))
    def test_osx_memdata_with_comma(self):
        """
        test osx memdata method when comma returns
        """

        def _cmd_side_effect(cmd):
            if "hw.memsize" in cmd:
                return "4294967296"
            elif "vm.swapusage" in cmd:
                return "total = 1024,00M  used = 160,75M  free = 863,25M  (encrypted)"

        with patch.dict(
            core.__salt__, {"cmd.run": MagicMock(side_effect=_cmd_side_effect)}
        ):
            ret = core._osx_memdata()
            assert ret["swap_total"] == 1024
            assert ret["mem_total"] == 4096

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/sbin/sysctl"))
    def test_osx_memdata(self):
        """
        test osx memdata
        """

        def _cmd_side_effect(cmd):
            if "hw.memsize" in cmd:
                return "4294967296"
            elif "vm.swapusage" in cmd:
                return "total = 0.00M  used = 0.00M  free = 0.00M  (encrypted)"

        with patch.dict(
            core.__salt__, {"cmd.run": MagicMock(side_effect=_cmd_side_effect)}
        ):
            ret = core._osx_memdata()
            assert ret["swap_total"] == 0
            assert ret["mem_total"] == 4096

    @skipIf(not core._DATEUTIL_TZ, "Missing dateutil.tz")
    def test_locale_info_tzname(self):
        # mock datetime.now().tzname()
        # cant just mock now because it is read only
        tzname = Mock(return_value="MDT_FAKE")
        now_ret_object = Mock(tzname=tzname)
        now = Mock(return_value=now_ret_object)
        datetime = Mock(now=now)

        with patch.object(core, "datetime", datetime=datetime) as datetime_module:
            with patch.object(
                core.dateutil.tz, "tzlocal", return_value=object
            ) as tzlocal:
                with patch.object(
                    salt.utils.platform, "is_proxy", return_value=False
                ) as is_proxy:
                    ret = core.locale_info()

                    tzname.assert_called_once_with()
                    self.assertEqual(len(now_ret_object.method_calls), 1)
                    now.assert_called_once_with(object)
                    self.assertEqual(len(datetime.method_calls), 1)
                    self.assertEqual(len(datetime_module.method_calls), 1)
                    tzlocal.assert_called_once_with()
                    is_proxy.assert_called_once_with()

                    self.assertEqual(ret["locale_info"]["timezone"], "MDT_FAKE")

    @skipIf(not core._DATEUTIL_TZ, "Missing dateutil.tz")
    def test_locale_info_unicode_error_tzname(self):
        # UnicodeDecodeError most have the default string encoding
        unicode_error = UnicodeDecodeError(str("fake"), b"\x00\x00", 1, 2, str("fake"))

        # mock datetime.now().tzname()
        # cant just mock now because it is read only
        tzname = Mock(return_value="MDT_FAKE")
        now_ret_object = Mock(tzname=tzname)
        now = Mock(return_value=now_ret_object)
        datetime = Mock(now=now)

        # mock tzname[0].decode()
        decode = Mock(return_value="CST_FAKE")
        tzname2 = (Mock(decode=decode,),)

        with patch.object(core, "datetime", datetime=datetime) as datetime_module:
            with patch.object(
                core.dateutil.tz, "tzlocal", side_effect=unicode_error
            ) as tzlocal:
                with patch.object(
                    salt.utils.platform, "is_proxy", return_value=False
                ) as is_proxy:
                    with patch.object(
                        core.salt.utils.platform, "is_windows", return_value=True
                    ) as is_windows:
                        with patch.object(core, "time", tzname=tzname2):
                            ret = core.locale_info()

                            tzname.assert_not_called()
                            self.assertEqual(len(now_ret_object.method_calls), 0)
                            now.assert_not_called()
                            self.assertEqual(len(datetime.method_calls), 0)
                            decode.assert_called_once_with("mbcs")
                            self.assertEqual(len(tzname2[0].method_calls), 1)
                            self.assertEqual(len(datetime_module.method_calls), 0)
                            tzlocal.assert_called_once_with()
                            is_proxy.assert_called_once_with()
                            is_windows.assert_called_once_with()

                            self.assertEqual(ret["locale_info"]["timezone"], "CST_FAKE")

    @skipIf(core._DATEUTIL_TZ, "Not Missing dateutil.tz")
    def test_locale_info_no_tz_tzname(self):
        with patch.object(
            salt.utils.platform, "is_proxy", return_value=False
        ) as is_proxy:
            with patch.object(
                core.salt.utils.platform, "is_windows", return_value=True
            ) as is_windows:
                ret = core.locale_info()
                is_proxy.assert_called_once_with()
                is_windows.assert_not_called()
                self.assertEqual(ret["locale_info"]["timezone"], "unknown")

    def test_cwd_exists(self):
        cwd_grain = core.cwd()

        self.assertIsInstance(cwd_grain, dict)
        self.assertTrue("cwd" in cwd_grain)
        self.assertEqual(cwd_grain["cwd"], os.getcwd())

    def test_cwd_is_cwd(self):
        cwd = os.getcwd()

        try:
            # change directory
            new_dir = os.path.split(cwd)[0]
            os.chdir(new_dir)

            cwd_grain = core.cwd()

            self.assertEqual(cwd_grain["cwd"], new_dir)
        finally:
            # change back to original directory
            os.chdir(cwd)

    def test_virtual_set_virtual_grain(self):
        osdata = {}

        (
            osdata["kernel"],
            osdata["nodename"],
            osdata["kernelrelease"],
            osdata["kernelversion"],
            osdata["cpuarch"],
            _,
        ) = platform.uname()

        with patch.dict(
            core.__salt__,
            {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.retcode": salt.modules.cmdmod.retcode,
                "smbios.get": salt.modules.smbios.get,
            },
        ):

            virtual_grains = core._virtual(osdata)

        self.assertIn("virtual", virtual_grains)

    def test_virtual_has_virtual_grain(self):
        osdata = {"virtual": "something"}

        (
            osdata["kernel"],
            osdata["nodename"],
            osdata["kernelrelease"],
            osdata["kernelversion"],
            osdata["cpuarch"],
            _,
        ) = platform.uname()

        with patch.dict(
            core.__salt__,
            {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.retcode": salt.modules.cmdmod.retcode,
                "smbios.get": salt.modules.smbios.get,
            },
        ):

            virtual_grains = core._virtual(osdata)

        self.assertIn("virtual", virtual_grains)
        self.assertNotEqual(virtual_grains["virtual"], "physical")

    @skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
    def test_windows_virtual_set_virtual_grain(self):
        osdata = {}

        (
            osdata["kernel"],
            osdata["nodename"],
            osdata["kernelrelease"],
            osdata["kernelversion"],
            osdata["cpuarch"],
            _,
        ) = platform.uname()

        with patch.dict(
            core.__salt__,
            {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.retcode": salt.modules.cmdmod.retcode,
                "smbios.get": salt.modules.smbios.get,
            },
        ):

            virtual_grains = core._windows_virtual(osdata)

        self.assertIn("virtual", virtual_grains)

    @skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
    def test_windows_virtual_has_virtual_grain(self):
        osdata = {"virtual": "something"}

        (
            osdata["kernel"],
            osdata["nodename"],
            osdata["kernelrelease"],
            osdata["kernelversion"],
            osdata["cpuarch"],
            _,
        ) = platform.uname()

        with patch.dict(
            core.__salt__,
            {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.retcode": salt.modules.cmdmod.retcode,
                "smbios.get": salt.modules.smbios.get,
            },
        ):

            virtual_grains = core._windows_virtual(osdata)

        self.assertIn("virtual", virtual_grains)
        self.assertNotEqual(virtual_grains["virtual"], "physical")

    @skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
    def test_osdata_virtual_key_win(self):
        with patch.dict(
            core.__salt__,
            {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.retcode": salt.modules.cmdmod.retcode,
                "smbios.get": salt.modules.smbios.get,
            },
        ):

            _windows_platform_data_ret = core.os_data()
            _windows_platform_data_ret["virtual"] = "something"

            with patch.object(
                core, "_windows_platform_data", return_value=_windows_platform_data_ret
            ) as _windows_platform_data:

                osdata_grains = core.os_data()
                _windows_platform_data.assert_called_once()

            self.assertIn("virtual", osdata_grains)
            self.assertNotEqual(osdata_grains["virtual"], "physical")

    @skipIf(salt.utils.platform.is_windows(), "System is Windows")
    def test_bsd_osfullname(self):
        """
        Test to ensure osfullname exists on *BSD systems
        """
        _path_exists_map = {}
        _path_isfile_map = {}
        _cmd_run_map = {
            "freebsd-version -u": "10.3-RELEASE",
            "/sbin/sysctl -n hw.physmem": "2121781248",
            "/sbin/sysctl -n vm.swap_total": "419430400",
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        path_isfile_mock = MagicMock(
            side_effect=lambda x: _path_isfile_map.get(x, False)
        )
        cmd_run_mock = MagicMock(side_effect=lambda x: _cmd_run_map[x])
        empty_mock = MagicMock(return_value={})

        mock_freebsd_uname = (
            "FreeBSD",
            "freebsd10.3-hostname-8148",
            "10.3-RELEASE",
            "FreeBSD 10.3-RELEASE #0 r297264: Fri Mar 25 02:10:02 UTC 2016     root@releng1.nyi.freebsd.org:/usr/obj/usr/src/sys/GENERIC",
            "amd64",
            "amd64",
        )

        with patch("platform.uname", MagicMock(return_value=mock_freebsd_uname)):
            with patch.object(
                salt.utils.platform, "is_linux", MagicMock(return_value=False)
            ):
                with patch.object(
                    salt.utils.platform, "is_freebsd", MagicMock(return_value=True)
                ):
                    # Skip the first if statement
                    with patch.object(
                        salt.utils.platform, "is_proxy", MagicMock(return_value=False)
                    ):
                        # Skip the init grain compilation (not pertinent)
                        with patch.object(os.path, "exists", path_exists_mock):
                            with patch("salt.utils.path.which") as mock:
                                mock.return_value = "/sbin/sysctl"
                                # Make a bunch of functions return empty dicts,
                                # we don't care about these grains for the
                                # purposes of this test.
                                with patch.object(
                                    core, "_bsd_cpudata", empty_mock
                                ), patch.object(
                                    core, "_hw_data", empty_mock
                                ), patch.object(
                                    core, "_virtual", empty_mock
                                ), patch.object(
                                    core, "_ps", empty_mock
                                ), patch.dict(
                                    core.__salt__, {"cmd.run": cmd_run_mock}
                                ):
                                    os_grains = core.os_data()

        self.assertIn("osfullname", os_grains)
        self.assertEqual(os_grains.get("osfullname"), "FreeBSD")

    def test_saltversioninfo(self):
        """
        test saltversioninfo core grain.
        """
        ret = core.saltversioninfo()
        info = ret["saltversioninfo"]
        assert isinstance(ret, dict)
        assert isinstance(info, list)
        try:
            assert len(info) == 1
        except AssertionError:
            # We have a minor version we need to test
            assert len(info) == 2
        assert all([x is not None for x in info])
        assert all([isinstance(x, int) for x in info])

    def test_path(self):
        comps = ["foo", "bar", "baz"]
        path = os.path.pathsep.join(comps)
        with patch.dict(os.environ, {"PATH": path}):
            result = core.path()
        assert result == {"path": path, "systempath": comps}, result

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    @patch("os.path.exists")
    @patch("salt.utils.platform.is_proxy")
    def test__hw_data_linux_empty(self, is_proxy, exists):
        is_proxy.return_value = False
        exists.return_value = True
        with patch("salt.utils.files.fopen", mock_open(read_data="")):
            self.assertEqual(
                core._hw_data({"kernel": "Linux"}),
                {
                    "biosreleasedate": "",
                    "biosversion": "",
                    "manufacturer": "",
                    "productname": "",
                    "serialnumber": "",
                    "uuid": "",
                },
            )

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    @patch("os.path.exists")
    @patch("salt.utils.platform.is_proxy")
    def test__hw_data_linux_unicode_error(self, is_proxy, exists):
        def _fopen(*args):
            class _File(object):
                def __enter__(self):
                    return self

                def __exit__(self, *args):
                    pass

                def read(self):
                    raise UnicodeDecodeError("enconding", b"", 1, 2, "reason")

            return _File()

        is_proxy.return_value = False
        exists.return_value = True
        with patch("salt.utils.files.fopen", _fopen):
            self.assertEqual(core._hw_data({"kernel": "Linux"}), {})

    @skipIf(not salt.utils.platform.is_linux(), "System is not Linux")
    def test_kernelparams_return(self):
        expectations = [
            (
                "BOOT_IMAGE=/vmlinuz-3.10.0-693.2.2.el7.x86_64",
                {
                    "kernelparams": [
                        ("BOOT_IMAGE", "/vmlinuz-3.10.0-693.2.2.el7.x86_64")
                    ]
                },
            ),
            (
                "root=/dev/mapper/centos_daemon-root",
                {"kernelparams": [("root", "/dev/mapper/centos_daemon-root")]},
            ),
            (
                "rhgb quiet ro",
                {"kernelparams": [("rhgb", None), ("quiet", None), ("ro", None)]},
            ),
            ('param="value1"', {"kernelparams": [("param", "value1")]}),
            (
                'param="value1 value2 value3"',
                {"kernelparams": [("param", "value1 value2 value3")]},
            ),
            (
                'param="value1 value2 value3" LANG="pl" ro',
                {
                    "kernelparams": [
                        ("param", "value1 value2 value3"),
                        ("LANG", "pl"),
                        ("ro", None),
                    ]
                },
            ),
            ("ipv6.disable=1", {"kernelparams": [("ipv6.disable", "1")]}),
            (
                'param="value1:value2:value3"',
                {"kernelparams": [("param", "value1:value2:value3")]},
            ),
            (
                'param="value1,value2,value3"',
                {"kernelparams": [("param", "value1,value2,value3")]},
            ),
            (
                'param="value1" param="value2" param="value3"',
                {
                    "kernelparams": [
                        ("param", "value1"),
                        ("param", "value2"),
                        ("param", "value3"),
                    ]
                },
            ),
        ]

        for cmdline, expectation in expectations:
            with patch("salt.utils.files.fopen", mock_open(read_data=cmdline)):
                self.assertEqual(core.kernelparams(), expectation)

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/sbin/lspci"))
    def test_linux_gpus(self):
        """
        Test GPU detection on Linux systems
        """

        def _cmd_side_effect(cmd):
            ret = ""
            for device in devices:
                ret += textwrap.dedent(
                    """
                                          Class:	{0}
                                          Vendor:	{1}
                                          Device:	{2}
                                          SVendor:	Evil Corp.
                                          SDevice:	Graphics XXL
                                          Rev:	c1
                                          NUMANode:	0"""
                ).format(*device)
                ret += "\n"
            return ret.strip()

        devices = [
            [
                "VGA compatible controller",
                "Advanced Micro Devices, Inc. [AMD/ATI]",
                "Vega [Radeon RX Vega]]",
                "amd",
            ],  # AMD
            [
                "Audio device",
                "Advanced Micro Devices, Inc. [AMD/ATI]",
                "Device aaf8",
                None,
            ],  # non-GPU device
            [
                "VGA compatible controller",
                "NVIDIA Corporation",
                "GK208 [GeForce GT 730]",
                "nvidia",
            ],  # Nvidia
            [
                "VGA compatible controller",
                "Intel Corporation",
                "Device 5912",
                "intel",
            ],  # Intel
            [
                "VGA compatible controller",
                "ATI Technologies Inc",
                "RC410 [Radeon Xpress 200M]",
                "ati",
            ],  # ATI
            [
                "3D controller",
                "NVIDIA Corporation",
                "GeForce GTX 950M",
                "nvidia",
            ],  # 3D controller
        ]
        with patch.dict(
            core.__salt__, {"cmd.run": MagicMock(side_effect=_cmd_side_effect)}
        ):
            ret = core._linux_gpu_data()["gpus"]
            count = 0
            for device in devices:
                if device[3] is None:
                    continue
                assert ret[count]["model"] == device[2]
                assert ret[count]["vendor"] == device[3]
                count += 1

    def test_get_server_id(self):
        expected = {"server_id": 94889706}
        with patch.dict(core.__opts__, {"id": "anid"}):
            assert core.get_server_id() == expected

        with patch.dict(core.__opts__, {"id": "otherid"}):
            assert core.get_server_id() != expected



# -*- coding: utf-8 -*-
"""
Direct call executor module
"""
from __future__ import absolute_import, print_function, unicode_literals


def execute(opts, data, func, args, kwargs):
    """
    Directly calls the given function with arguments
    """
    return func(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

import logging

# Import Salt Libs
import salt.states.user as user

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class UserTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the user state
    """

    def setup_loader_modules(self):
        return {user: {}}

    def test_present(self):
        """
            Test to ensure that the named user is present with
            the specified properties
        """
        ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
        mock_false = MagicMock(return_value=False)
        mock_empty_list = MagicMock(return_value=[])
        with patch.dict(user.__grains__, {"kernel": "Linux"}):
            with patch.dict(
                user.__salt__,
                {
                    "group.info": mock_false,
                    "user.info": mock_empty_list,
                    "user.chkey": mock_empty_list,
                    "user.add": mock_false,
                },
            ):
                ret.update(
                    {"comment": "The following group(s) are" " not present: salt"}
                )
                self.assertDictEqual(user.present("salt", groups=["salt"]), ret)

                mock_false = MagicMock(
                    side_effect=[
                        {"key": "value"},
                        {"key": "value"},
                        {"key": "value"},
                        False,
                        False,
                    ]
                )
                with patch.object(user, "_changes", mock_false):
                    with patch.dict(user.__opts__, {"test": True}):
                        ret.update(
                            {
                                "comment": "The following user attributes are set "
                                "to be changed:\n"
                                "key: value\n",
                                "result": None,
                            }
                        )
                        self.assertDictEqual(user.present("salt"), ret)

                    with patch.dict(user.__opts__, {"test": False}):
                        # pylint: disable=repr-flag-used-in-string
                        comment = "These values could not be changed: {0!r}".format(
                            {"key": "value"}
                        )
                        # pylint: enable=repr-flag-used-in-string
                        ret.update({"comment": comment, "result": False})
                        self.assertDictEqual(user.present("salt"), ret)

                        with patch.dict(user.__opts__, {"test": True}):
                            ret.update(
                                {
                                    "comment": "User salt set to" " be added",
                                    "result": None,
                                }
                            )
                            self.assertDictEqual(user.present("salt"), ret)

                        with patch.dict(user.__opts__, {"test": False}):
                            ret.update(
                                {
                                    "comment": "Failed to create new" " user salt",
                                    "result": False,
                                }
                            )
                            self.assertDictEqual(user.present("salt"), ret)

    def test_present_invalid_uid_change(self):
        mock_info = MagicMock(
            side_effect=[
                {
                    "uid": 5000,
                    "gid": 5000,
                    "groups": ["foo"],
                    "home": "/home/foo",
                    "fullname": "Foo Bar",
                }
            ]
        )
        dunder_salt = {
            "user.info": mock_info,
            "file.group_to_gid": MagicMock(side_effect=["foo"]),
            "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
        }
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
            user.__salt__, dunder_salt
        ):
            ret = user.present("foo", uid=5001)
            # State should have failed
            self.assertFalse(ret["result"])
            # Only one of uid/gid should have been flagged in the comment
            self.assertEqual(ret["comment"].count("not permitted"), 1)

    def test_present_invalid_gid_change(self):
        mock_info = MagicMock(
            side_effect=[
                {
                    "uid": 5000,
                    "gid": 5000,
                    "groups": ["foo"],
                    "home": "/home/foo",
                    "fullname": "Foo Bar",
                }
            ]
        )
        dunder_salt = {
            "user.info": mock_info,
            "file.group_to_gid": MagicMock(side_effect=["foo"]),
            "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
        }
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
            user.__salt__, dunder_salt
        ):
            ret = user.present("foo", gid=5001)
            # State should have failed
            self.assertFalse(ret["result"])
            # Only one of uid/gid should have been flagged in the comment
            self.assertEqual(ret["comment"].count("not permitted"), 1)

    def test_present_invalid_uid_gid_change(self):
        mock_info = MagicMock(
            side_effect=[
                {
                    "uid": 5000,
                    "gid": 5000,
                    "groups": ["foo"],
                    "home": "/home/foo",
                    "fullname": "Foo Bar",
                }
            ]
        )
        dunder_salt = {
            "user.info": mock_info,
            "file.group_to_gid": MagicMock(side_effect=["foo"]),
            "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
        }
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
            user.__salt__, dunder_salt
        ):
            ret = user.present("foo", uid=5001, gid=5001)
            # State should have failed
            self.assertFalse(ret["result"])
            # Both the uid and gid should have been flagged in the comment
            self.assertEqual(ret["comment"].count("not permitted"), 2)

    def test_present_uid_gid_change(self):
        before = {
            "uid": 5000,
            "gid": 5000,
            "groups": ["foo"],
            "home": "/home/foo",
            "fullname": "Foo Bar",
        }
        after = {
            "uid": 5001,
            "gid": 5001,
            "groups": ["othergroup"],
            "home": "/home/foo",
            "fullname": "Foo Bar",
        }
        # user.info should be called 4 times. Once the first time that
        # _changes() is called, once before and after changes are applied (to
        # get the before/after for the changes dict, and one last time to
        # confirm that no changes still need to be made.
        mock_info = MagicMock(side_effect=[before, before, after, after])
        mock_group_to_gid = MagicMock(side_effect=["foo", "othergroup"])
        mock_gid_to_group = MagicMock(side_effect=[5000, 5000, 5001, 5001])
        dunder_salt = {
            "user.info": mock_info,
            "user.chuid": Mock(),
            "user.chgid": Mock(),
            "file.group_to_gid": mock_group_to_gid,
            "file.gid_to_group": mock_gid_to_group,
        }
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
            user.__salt__, dunder_salt
        ), patch.dict(user.__opts__, {"test": False}), patch(
            "os.path.isdir", MagicMock(return_value=True)
        ):
            ret = user.present(
                "foo", uid=5001, gid=5001, allow_uid_change=True, allow_gid_change=True
            )
            self.assertEqual(
                ret,
                {
                    "comment": "Updated user foo",
                    "changes": {"gid": 5001, "uid": 5001, "groups": ["othergroup"]},
                    "name": "foo",
                    "result": True,
                },
            )

    def test_absent(self):
        """
            Test to ensure that the named user is absent
        """
        ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}
        mock = MagicMock(side_effect=[True, True, False])
        mock1 = MagicMock(return_value=False)
        with patch.dict(
            user.__salt__,
            {"user.info": mock, "user.delete": mock1, "group.info": mock1},
        ):
            with patch.dict(user.__opts__, {"test": True}):
                ret.update({"comment": "User salt set for removal"})
                self.assertDictEqual(user.absent("salt"), ret)

            with patch.dict(user.__opts__, {"test": False}):
                ret.update({"comment": "Failed to remove user salt", "result": False})
                self.assertDictEqual(user.absent("salt"), ret)

            ret.update({"comment": "User salt is not present", "result": True})
            self.assertDictEqual(user.absent("salt"), ret)

    def test_changes(self):
        """
        Test salt.states.user._changes
        """
        mock_info = MagicMock(
            return_value={
                "uid": 5000,
                "gid": 5000,
                "groups": ["foo"],
                "home": "/home/foo",
                "fullname": "Foo Bar",
            }
        )
        shadow_info = MagicMock(
            return_value={"min": 2, "max": 88888, "inact": 77, "warn": 14}
        )
        shadow_hash = MagicMock(return_value="abcd")
        dunder_salt = {
            "user.info": mock_info,
            "shadow.info": shadow_info,
            "shadow.default_hash": shadow_hash,
            "file.group_to_gid": MagicMock(side_effect=["foo"]),
            "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
        }

        def mock_exists(*args):
            return True

        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
            user.__salt__, dunder_salt
        ), patch.dict(user.__opts__, {"test": False}), patch(
            "os.path.isdir", mock_exists
        ):
            ret = user._changes("foo", maxdays=999999, inactdays=0, warndays=7)
            assert ret == {
                "maxdays": 999999,
                "mindays": 0,
                "fullname": "",
                "warndays": 7,
                "inactdays": 0,
            }



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.redismod as redismod

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RedismodTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.redismod
    """

    def setup_loader_modules(self):
        return {redismod: {}}

    # 'string' function tests: 1

    def test_string(self):
        """
        Test to ensure that the key exists in redis with the value specified.
        """
        name = "key_in_redis"
        value = "string data"

        ret = {
            "name": name,
            "changes": {},
            "result": True,
            "comment": "Key already set to defined value",
        }

        mock = MagicMock(return_value=value)
        with patch.dict(redismod.__salt__, {"redis.get_key": mock}):
            self.assertDictEqual(redismod.string(name, value), ret)

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to ensure key absent from redis.
        """
        name = "key_in_redis"

        ret = {"name": name, "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[False, True, True])
        mock_t = MagicMock(return_value=False)
        with patch.dict(
            redismod.__salt__, {"redis.exists": mock, "redis.delete": mock_t}
        ):
            comt = "`keys` not formed as a list type"
            ret.update({"comment": comt, "result": False})
            self.assertDictEqual(redismod.absent(name, "key"), ret)

            comt = "Key(s) specified already absent"
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(redismod.absent(name, ["key"]), ret)

            comt = "Keys deleted"
            ret.update({"comment": comt, "changes": {"deleted": ["key"]}})
            self.assertDictEqual(redismod.absent(name, ["key"]), ret)

            comt = "Key deleted"
            ret.update({"comment": comt, "changes": {"deleted": ["key_in_redis"]}})
            self.assertDictEqual(redismod.absent(name), ret)



# -*- coding: utf-8 -*-
"""

    integration.reactor.reactor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test Salt's reactor system
"""

from __future__ import absolute_import

import signal

import pytest
import salt.utils.event
import salt.utils.reactor
from tests.support.case import ShellCase
from tests.support.mixins import SaltMinionEventAssertsMixin
from tests.support.unit import skipIf


class TimeoutException(Exception):
    pass


@pytest.mark.windows_whitelisted
class ReactorTest(SaltMinionEventAssertsMixin, ShellCase):
    """
    Test Salt's reactor system
    """

    def setUp(self):
        self.timeout = 30

    def get_event(self, class_type="master"):
        return salt.utils.event.get_event(
            class_type,
            sock_dir=self.master_opts["sock_dir"],
            transport=self.master_opts["transport"],
            keep_loop=True,
            opts=self.master_opts,
        )

    def fire_event(self, tag, data):
        event = self.get_event()
        event.fire_event(tag, data)

    def alarm_handler(self, signal, frame):
        raise TimeoutException("Timeout of {0} seconds reached".format(self.timeout))

    def test_ping_reaction(self):
        """
        Fire an event on the master and ensure
        that it pings the minion
        """
        # Create event bus connection
        e = salt.utils.event.get_event(
            "minion", sock_dir=self.minion_opts["sock_dir"], opts=self.minion_opts
        )

        e.fire_event({"a": "b"}, "/test_event")

        self.assertMinionEventReceived({"a": "b"}, timeout=30)

    @skipIf(salt.utils.platform.is_windows(), "no sigalarm on windows")
    def test_reactor_reaction(self):
        """
        Fire an event on the master and ensure
        The reactor event responds
        """
        signal.signal(signal.SIGALRM, self.alarm_handler)
        signal.alarm(self.timeout)

        master_event = self.get_event()
        master_event.fire_event({"id": "minion"}, "salt/test/reactor")

        try:
            while True:
                event = master_event.get_event(full=True)

                if event is None:
                    continue

                if event.get("tag") == "test_reaction":
                    self.assertTrue(event["data"]["test_reaction"])
                    break
        finally:
            signal.alarm(0)

    @skipIf(salt.utils.platform.is_windows(), "no sigalarm on windows")
    def test_reactor_is_leader(self):
        """
        when leader is set to false reactor should timeout/not do anything
        """
        # by default reactor should be leader
        ret = self.run_run_plus("reactor.is_leader")
        self.assertTrue(ret["return"])

        # make reactor not leader
        self.run_run_plus("reactor.set_leader", False)
        ret = self.run_run_plus("reactor.is_leader")
        self.assertFalse(ret["return"])

        signal.signal(signal.SIGALRM, self.alarm_handler)
        signal.alarm(self.timeout)

        try:
            master_event = self.get_event()
            self.fire_event({"id": "minion"}, "salt/test/reactor")

            while True:
                event = master_event.get_event(full=True)

                if event is None:
                    continue

                if event.get("tag") == "test_reaction":
                    # if we reach this point, the test is a failure
                    self.assertTrue(False)  # pylint: disable=redundant-unittest-assert
                    break
        except TimeoutException as exc:
            self.assertTrue("Timeout" in str(exc))
        finally:
            signal.alarm(0)

        # make reactor leader again
        self.run_run_plus("reactor.set_leader", True)
        ret = self.run_run_plus("reactor.is_leader")
        self.assertTrue(ret["return"])

        # trigger a reaction
        signal.alarm(self.timeout)

        try:
            master_event = self.get_event()
            self.fire_event({"id": "minion"}, "salt/test/reactor")

            while True:
                event = master_event.get_event(full=True)

                if event is None:
                    continue

                if event.get("tag") == "test_reaction":
                    self.assertTrue(event["data"]["test_reaction"])
                    break
        finally:
            signal.alarm(0)



# -*- coding: utf-8 -*-
"""
Management of Zabbix hosts.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>


"""
from __future__ import absolute_import, print_function, unicode_literals

from copy import deepcopy

import salt.utils.json
from salt.ext import six


def __virtual__():
    """
    Only make these states available if Zabbix module is available.
    """
    if "zabbix.host_create" in __salt__:
        return True
    return (False, "zabbix module could not be loaded")


def present(host, groups, interfaces, **kwargs):
    """
    Ensures that the host exists, eventually creates new host.
    NOTE: please use argument visible_name instead of name to not mess with name from salt sls. This function accepts
    all standard host properties: keyword argument names differ depending on your zabbix version, see:
    https://www.zabbix.com/documentation/2.4/manual/api/reference/host/object#host

    .. versionadded:: 2016.3.0

    :param host: technical name of the host
    :param groups: groupids of host groups to add the host to
    :param interfaces: interfaces to be created for the host
    :param proxy_host: Optional proxy name or proxyid to monitor host
    :param inventory: Optional list of inventory names and values
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)
    :param visible_name: Optional - string with visible name of the host, use 'visible_name' instead of 'name' \
    parameter to not mess with value supplied from Salt sls file.

    .. code-block:: yaml

        create_test_host:
            zabbix_host.present:
                - host: TestHostWithInterfaces
                - proxy_host: 12345
                - groups:
                    - 5
                    - 6
                    - 7
                - interfaces:
                    - test1.example.com:
                        - ip: '192.168.1.8'
                        - type: 'Agent'
                        - port: 92
                    - testing2_create:
                        - ip: '192.168.1.9'
                        - dns: 'test2.example.com'
                        - type: 'agent'
                        - main: false
                    - testovaci1_ipmi:
                        - ip: '192.168.100.111'
                        - type: 'ipmi'
                - inventory:
                    - alias: some alias
                    - asset_tag: jlm3937


    """
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": host, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_host_created = "Host {0} created.".format(host)
    comment_host_updated = "Host {0} updated.".format(host)
    comment_host_notcreated = "Unable to create host: {0}. ".format(host)
    comment_host_exists = "Host {0} already exists.".format(host)
    changes_host_created = {
        host: {
            "old": "Host {0} does not exist.".format(host),
            "new": "Host {0} created.".format(host),
        }
    }

    def _interface_format(interfaces_data):
        """
        Formats interfaces from SLS file into valid JSON usable for zabbix API.
        Completes JSON with default values.

        :param interfaces_data: list of interfaces data from SLS file

        """

        if not interfaces_data:
            return list()

        interface_attrs = ("ip", "dns", "main", "type", "useip", "port")
        interfaces_json = salt.utils.json.loads(salt.utils.json.dumps(interfaces_data))
        interfaces_dict = dict()

        for interface in interfaces_json:
            for intf in interface:
                intf_name = intf
                interfaces_dict[intf_name] = dict()
                for intf_val in interface[intf]:
                    for key, value in intf_val.items():
                        if key in interface_attrs:
                            interfaces_dict[intf_name][key] = value

        interfaces_list = list()
        interface_ports = {
            "agent": ["1", "10050"],
            "snmp": ["2", "161"],
            "ipmi": ["3", "623"],
            "jmx": ["4", "12345"],
        }

        for key, value in interfaces_dict.items():
            # Load interface values or default values
            interface_type = interface_ports[value["type"].lower()][0]
            main = (
                "1"
                if six.text_type(value.get("main", "true")).lower() == "true"
                else "0"
            )
            useip = (
                "1"
                if six.text_type(value.get("useip", "true")).lower() == "true"
                else "0"
            )
            interface_ip = value.get("ip", "")
            dns = value.get("dns", key)
            port = six.text_type(
                value.get("port", interface_ports[value["type"].lower()][1])
            )

            interfaces_list.append(
                {
                    "type": interface_type,
                    "main": main,
                    "useip": useip,
                    "ip": interface_ip,
                    "dns": dns,
                    "port": port,
                }
            )

        interfaces_list_sorted = sorted(
            interfaces_list, key=lambda k: k["main"], reverse=True
        )

        return interfaces_list_sorted

    interfaces_formated = _interface_format(interfaces)

    # Ensure groups are all groupid
    groupids = []
    for group in groups:
        if isinstance(group, six.string_types):
            groupid = __salt__["zabbix.hostgroup_get"](name=group, **connection_args)
            try:
                groupids.append(int(groupid[0]["groupid"]))
            except TypeError:
                ret["comment"] = "Invalid group {0}".format(group)
                return ret
        else:
            groupids.append(group)
    groups = groupids

    # Get and validate proxyid
    proxy_hostid = "0"
    if "proxy_host" in kwargs:
        # Test if proxy_host given as name
        if isinstance(kwargs["proxy_host"], six.string_types):
            try:
                proxy_hostid = __salt__["zabbix.run_query"](
                    "proxy.get",
                    {
                        "output": "proxyid",
                        "selectInterface": "extend",
                        "filter": {"host": "{0}".format(kwargs["proxy_host"])},
                    },
                    **connection_args
                )[0]["proxyid"]
            except TypeError:
                ret["comment"] = "Invalid proxy_host {0}".format(kwargs["proxy_host"])
                return ret
        # Otherwise lookup proxy_host as proxyid
        else:
            try:
                proxy_hostid = __salt__["zabbix.run_query"](
                    "proxy.get",
                    {
                        "proxyids": "{0}".format(kwargs["proxy_host"]),
                        "output": "proxyid",
                    },
                    **connection_args
                )[0]["proxyid"]
            except TypeError:
                ret["comment"] = "Invalid proxy_host {0}".format(kwargs["proxy_host"])
                return ret

    if "inventory" not in kwargs:
        inventory = {}
    else:
        inventory = kwargs["inventory"]
    if inventory is None:
        inventory = {}
    # Create dict of requested inventory items
    new_inventory = {}
    for inv_item in inventory:
        for k, v in inv_item.items():
            new_inventory[k] = str(v)

    host_exists = __salt__["zabbix.host_exists"](host, **connection_args)

    if host_exists:
        host = __salt__["zabbix.host_get"](host=host, **connection_args)[0]
        hostid = host["hostid"]

        update_proxy = False
        update_hostgroups = False
        update_interfaces = False
        update_inventory = False

        cur_proxy_hostid = host["proxy_hostid"]
        if proxy_hostid != cur_proxy_hostid:
            update_proxy = True

        hostgroups = __salt__["zabbix.hostgroup_get"](hostids=hostid, **connection_args)
        cur_hostgroups = list()

        for hostgroup in hostgroups:
            cur_hostgroups.append(int(hostgroup["groupid"]))

        if set(groups) != set(cur_hostgroups):
            update_hostgroups = True

        hostinterfaces = __salt__["zabbix.hostinterface_get"](
            hostids=hostid, **connection_args
        )

        if hostinterfaces:
            hostinterfaces = sorted(hostinterfaces, key=lambda k: k["main"])
            hostinterfaces_copy = deepcopy(hostinterfaces)
            for hostintf in hostinterfaces_copy:
                hostintf.pop("interfaceid")
                hostintf.pop("bulk")
                hostintf.pop("hostid")
            interface_diff = [
                x for x in interfaces_formated if x not in hostinterfaces_copy
            ] + [y for y in hostinterfaces_copy if y not in interfaces_formated]
            if interface_diff:
                update_interfaces = True

        elif not hostinterfaces and interfaces:
            update_interfaces = True

        cur_inventory = __salt__["zabbix.host_inventory_get"](
            hostids=hostid, **connection_args
        )
        if cur_inventory:
            # Remove blank inventory items
            cur_inventory = {k: v for k, v in cur_inventory.items() if v}
            # Remove persistent inventory keys for comparison
            cur_inventory.pop("hostid", None)
            cur_inventory.pop("inventory_mode", None)

        if new_inventory and not cur_inventory:
            update_inventory = True
        elif set(cur_inventory) != set(new_inventory):
            update_inventory = True

    # Dry run, test=true mode
    if __opts__["test"]:
        if host_exists:
            if (
                update_hostgroups
                or update_interfaces
                or update_proxy
                or update_inventory
            ):
                ret["result"] = None
                ret["comment"] = comment_host_updated
            else:
                ret["result"] = True
                ret["comment"] = comment_host_exists
        else:
            ret["result"] = None
            ret["comment"] = comment_host_created
            ret["changes"] = changes_host_created
        return ret

    error = []

    if host_exists:
        ret["result"] = True
        if update_hostgroups or update_interfaces or update_proxy or update_inventory:

            if update_inventory:
                # combine connection_args, inventory, and clear_old
                sum_kwargs = dict(new_inventory)
                sum_kwargs.update(connection_args)
                sum_kwargs["clear_old"] = True

                hostupdate = __salt__["zabbix.host_inventory_set"](hostid, **sum_kwargs)
                ret["changes"]["inventory"] = str(new_inventory)
                if "error" in hostupdate:
                    error.append(hostupdate["error"])
            if update_proxy:
                hostupdate = __salt__["zabbix.host_update"](
                    hostid, proxy_hostid=proxy_hostid, **connection_args
                )
                ret["changes"]["proxy_hostid"] = six.text_type(proxy_hostid)
                if "error" in hostupdate:
                    error.append(hostupdate["error"])
            if update_hostgroups:
                hostupdate = __salt__["zabbix.host_update"](
                    hostid, groups=groups, **connection_args
                )
                ret["changes"]["groups"] = six.text_type(groups)
                if "error" in hostupdate:
                    error.append(hostupdate["error"])
            if update_interfaces:
                if hostinterfaces:
                    for interface in hostinterfaces:
                        __salt__["zabbix.hostinterface_delete"](
                            interfaceids=interface["interfaceid"], **connection_args
                        )

                hostid = __salt__["zabbix.host_get"](name=host, **connection_args)[0][
                    "hostid"
                ]

                for interface in interfaces_formated:
                    updatedint = __salt__["zabbix.hostinterface_create"](
                        hostid=hostid,
                        ip=interface["ip"],
                        dns=interface["dns"],
                        main=interface["main"],
                        type=interface["type"],
                        useip=interface["useip"],
                        port=interface["port"],
                        **connection_args
                    )

                    if "error" in updatedint:
                        error.append(updatedint["error"])

                ret["changes"]["interfaces"] = six.text_type(interfaces_formated)

            ret["comment"] = comment_host_updated

        else:
            ret["comment"] = comment_host_exists
    else:
        host_create = __salt__["zabbix.host_create"](
            host,
            groups,
            interfaces_formated,
            proxy_hostid=proxy_hostid,
            inventory=new_inventory,
            **connection_args
        )

        if "error" not in host_create:
            ret["result"] = True
            ret["comment"] = comment_host_created
            ret["changes"] = changes_host_created
        else:
            ret["result"] = False
            ret["comment"] = comment_host_notcreated + six.text_type(
                host_create["error"]
            )

    # error detected
    if error:
        ret["changes"] = {}
        ret["result"] = False
        ret["comment"] = six.text_type(error)

    return ret


def absent(name, **kwargs):
    """
    Ensures that the host does not exists, eventually deletes host.

    .. versionadded:: 2016.3.0

    :param: name: technical name of the host
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        TestHostWithInterfaces:
            zabbix_host.absent

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_host_deleted = "Host {0} deleted.".format(name)
    comment_host_notdeleted = "Unable to delete host: {0}. ".format(name)
    comment_host_notexists = "Host {0} does not exist.".format(name)
    changes_host_deleted = {
        name: {
            "old": "Host {0} exists.".format(name),
            "new": "Host {0} deleted.".format(name),
        }
    }
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    host_exists = __salt__["zabbix.host_exists"](name, **connection_args)

    # Dry run, test=true mode
    if __opts__["test"]:
        if not host_exists:
            ret["result"] = True
            ret["comment"] = comment_host_notexists
        else:
            ret["result"] = None
            ret["comment"] = comment_host_deleted
        return ret

    host_get = __salt__["zabbix.host_get"](name, **connection_args)

    if not host_get:
        ret["result"] = True
        ret["comment"] = comment_host_notexists
    else:
        try:
            hostid = host_get[0]["hostid"]
            host_delete = __salt__["zabbix.host_delete"](hostid, **connection_args)
        except KeyError:
            host_delete = False

        if host_delete and "error" not in host_delete:
            ret["result"] = True
            ret["comment"] = comment_host_deleted
            ret["changes"] = changes_host_deleted
        else:
            ret["result"] = False
            ret["comment"] = comment_host_notdeleted + six.text_type(
                host_delete["error"]
            )

    return ret


def assign_templates(host, templates, **kwargs):
    """
    Ensures that templates are assigned to the host.

    .. versionadded:: 2017.7.0

    :param host: technical name of the host
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        add_zabbix_templates_to_host:
            zabbix_host.assign_templates:
                - host: TestHost
                - templates:
                    - "Template OS Linux"
                    - "Template App MySQL"

    """
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": host, "changes": {}, "result": False, "comment": ""}

    # Set comments
    comment_host_templates_updated = "Templates updated."
    comment_host_templ_notupdated = "Unable to update templates on host: {0}.".format(
        host
    )
    comment_host_templates_in_sync = "Templates already synced."

    update_host_templates = False
    curr_template_ids = list()
    requested_template_ids = list()
    hostid = ""

    host_exists = __salt__["zabbix.host_exists"](host, **connection_args)

    # Fail out if host does not exist
    if not host_exists:
        ret["result"] = False
        ret["comment"] = comment_host_templ_notupdated
        return ret

    host_info = __salt__["zabbix.host_get"](host=host, **connection_args)[0]
    hostid = host_info["hostid"]

    if not templates:
        templates = list()

    # Get current templateids for host
    host_templates = __salt__["zabbix.host_get"](
        hostids=hostid,
        output='[{"hostid"}]',
        selectParentTemplates='["templateid"]',
        **connection_args
    )
    for template_id in host_templates[0]["parentTemplates"]:
        curr_template_ids.append(template_id["templateid"])

    # Get requested templateids
    for template in templates:
        try:
            template_id = __salt__["zabbix.template_get"](
                host=template, **connection_args
            )[0]["templateid"]
            requested_template_ids.append(template_id)
        except TypeError:
            ret["result"] = False
            ret["comment"] = "Unable to find template: {0}.".format(template)
            return ret

    # remove any duplications
    requested_template_ids = list(set(requested_template_ids))

    if set(curr_template_ids) != set(requested_template_ids):
        update_host_templates = True

    # Set change output
    changes_host_templates_modified = {
        host: {
            "old": "Host templates: " + ", ".join(curr_template_ids),
            "new": "Host templates: " + ", ".join(requested_template_ids),
        }
    }

    # Dry run, test=true mode
    if __opts__["test"]:
        if update_host_templates:
            ret["result"] = None
            ret["comment"] = comment_host_templates_updated
        else:
            ret["result"] = True
            ret["comment"] = comment_host_templates_in_sync
        return ret

    # Attempt to perform update
    ret["result"] = True
    if update_host_templates:
        update_output = __salt__["zabbix.host_update"](
            hostid, templates=(requested_template_ids), **connection_args
        )
        if update_output is False:
            ret["result"] = False
            ret["comment"] = comment_host_templ_notupdated
            return ret
        ret["comment"] = comment_host_templates_updated
        ret["changes"] = changes_host_templates_modified
    else:
        ret["comment"] = comment_host_templates_in_sync

    return ret



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import pytest
from tests.support.case import ModuleCase
from tests.support.helpers import (
    destructiveTest,
    random_string,
    requires_system_grains,
    runs_on,
    skip_if_not_root,
    slowTest,
)


@destructiveTest
@skip_if_not_root
@runs_on(kernel="Linux")
class UseraddModuleTestLinux(ModuleCase):
    @requires_system_grains
    @slowTest
    def test_groups_includes_primary(self, grains):
        # Let's create a user, which usually creates the group matching the
        # name
        uname = random_string("RS-", lowercase=False)
        if self.run_function("user.add", [uname]) is not True:
            # Skip because creating is not what we're testing here
            self.run_function("user.delete", [uname, True, True])
            self.skipTest("Failed to create user")

        try:
            uinfo = self.run_function("user.info", [uname])
            if grains["os_family"] in ("Suse",):
                self.assertIn("users", uinfo["groups"])
            else:
                self.assertIn(uname, uinfo["groups"])

            # This uid is available, store it
            uid = uinfo["uid"]

            self.run_function("user.delete", [uname, True, True])

            # Now, a weird group id
            gname = random_string("RS-", lowercase=False)
            if self.run_function("group.add", [gname]) is not True:
                self.run_function("group.delete", [gname, True, True])
                self.skipTest("Failed to create group")

            ginfo = self.run_function("group.info", [gname])

            # And create the user with that gid
            if self.run_function("user.add", [uname, uid, ginfo["gid"]]) is False:
                # Skip because creating is not what we're testing here
                self.run_function("user.delete", [uname, True, True])
                self.skipTest("Failed to create user")

            uinfo = self.run_function("user.info", [uname])
            self.assertIn(gname, uinfo["groups"])

        except AssertionError:
            self.run_function("user.delete", [uname, True, True])
            raise

    @slowTest
    def test_user_primary_group(self):
        """
        Tests the primary_group function
        """
        name = "saltyuser"

        # Create a user to test primary group function
        if self.run_function("user.add", [name]) is not True:
            self.run_function("user.delete", [name])
            self.skipTest("Failed to create a user")

        try:
            # Test useradd.primary_group
            primary_group = self.run_function("user.primary_group", [name])
            uid_info = self.run_function("user.info", [name])
            self.assertIn(primary_group, uid_info["groups"])

        except Exception:  # pylint: disable=broad-except
            self.run_function("user.delete", [name])
            raise


@destructiveTest
@skip_if_not_root
@runs_on(kernel="Windows")
@pytest.mark.windows_whitelisted
class UseraddModuleTestWindows(ModuleCase):
    def setUp(self):
        self.user_name = random_string("RS-", lowercase=False)
        self.group_name = random_string("RS-", lowercase=False)

    def tearDown(self):
        self.run_function("user.delete", [self.user_name, True, True])
        self.run_function("group.delete", [self.group_name])

    def _add_user(self):
        """
        helper class to add user
        """
        if self.run_function("user.add", [self.user_name]) is False:
            # Skip because creating is not what we're testing here
            self.skipTest("Failed to create user")

    def _add_group(self):
        """
        helper class to add group
        """
        if self.run_function("group.add", [self.group_name]) is False:
            # Skip because creating is not what we're testing here
            self.skipTest("Failed to create group")

    @slowTest
    def test_add_user(self):
        """
        Test adding a user
        """
        self._add_user()
        user_list = self.run_function("user.list_users")
        self.assertIn(self.user_name, user_list)

    @slowTest
    def test_add_group(self):
        """
        Test adding a user
        """
        self._add_group()
        group_list = self.run_function("group.list_groups")
        self.assertIn(self.group_name, group_list)

    @slowTest
    def test_add_user_to_group(self):
        """
        Test adding a user to a group
        """
        self._add_group()
        # And create the user as a member of that group
        self.run_function("user.add", [self.user_name], groups=self.group_name)
        user_info = self.run_function("user.info", [self.user_name])
        self.assertIn(self.group_name, user_info["groups"])

    @slowTest
    def test_add_user_addgroup(self):
        """
        Test adding a user to a group with groupadd
        """
        self._add_group()
        self._add_user()
        self.run_function("user.addgroup", [self.user_name, self.group_name])
        info = self.run_function("user.info", [self.user_name])
        self.assertEqual(info["groups"], [self.group_name])

    @slowTest
    def test_user_chhome(self):
        """
        Test changing a users home dir
        """
        self._add_user()
        user_dir = r"c:\salt"
        self.run_function("user.chhome", [self.user_name, user_dir])
        info = self.run_function("user.info", [self.user_name])
        self.assertEqual(info["home"], user_dir)

    @slowTest
    def test_user_chprofile(self):
        """
        Test changing a users profile
        """
        self._add_user()
        config = r"c:\salt\config"
        self.run_function("user.chprofile", [self.user_name, config])
        info = self.run_function("user.info", [self.user_name])
        self.assertEqual(info["profile"], config)

    @slowTest
    def test_user_chfullname(self):
        """
        Test changing a users fullname
        """
        self._add_user()
        name = "Salt Test"
        self.run_function("user.chfullname", [self.user_name, name])
        info = self.run_function("user.info", [self.user_name])
        self.assertEqual(info["fullname"], name)

    @slowTest
    def test_user_delete(self):
        """
        Test deleting a user
        """
        self._add_user()
        self.assertTrue(self.run_function("user.info", [self.user_name])["active"])
        self.run_function("user.delete", [self.user_name])
        self.assertEqual({}, self.run_function("user.info", [self.user_name]))

    @slowTest
    def test_user_removegroup(self):
        """
        Test removing a group
        """
        self._add_user()
        self._add_group()
        self.run_function("user.addgroup", [self.user_name, self.group_name])
        self.assertIn(
            self.group_name, self.run_function("user.list_groups", [self.user_name])
        )
        self.run_function("user.removegroup", [self.user_name, self.group_name])
        self.assertNotIn(
            self.group_name, self.run_function("user.list_groups", [self.user_name])
        )

    @slowTest
    def test_user_rename(self):
        """
        Test changing a users name
        """
        self._add_user()
        name = "newuser"
        self.run_function("user.rename", [self.user_name, name])
        info = self.run_function("user.info", [name])
        self.assertTrue(info["active"])

        # delete new user
        self.run_function("user.delete", [name, True, True])

    @slowTest
    def test_user_setpassword(self):
        """
        Test setting a password
        """
        self._add_user()
        passwd = "sup3rs3cr3T!"
        self.assertTrue(self.run_function("user.setpassword", [self.user_name, passwd]))



# -*- coding: utf-8 -*-
"""
    :codeauthor: :email:`Jakub Sliva <jakub.sliva@ultimum.io>`
"""

# Import Python Libs
from __future__ import absolute_import, unicode_literals

import salt.states.zabbix_valuemap as zabbix_valuemap

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

INPUT_PARAMS = {
    "mappings": [
        {"newvalue": "OK", "value": "0h"},
        {"newvalue": "Failure", "value": "1"},
    ],
    "name": "Server HP Health",
}

EXISTING_OBJ = [
    {
        "valuemapid": "21",
        "name": "Server HP Health",
        "mappings": [
            {"newvalue": "OK", "value": "0h"},
            {"newvalue": "Failure", "value": "1"},
        ],
    }
]

EXISTING_OBJ_DIFF = {
    "valuemapid": "21",
    "name": "Server HP Health",
    "mappings": [
        {"newvalue": "OK", "value": "0h"},
        {"newvalue": "Failure", "value": "1"},
        {"newvalue": "some", "value": "2"},
    ],
}

DIFF_PARAMS = {
    "valuemapid": "21",
    "mappings": [
        {"newvalue": "OK", "value": "0h"},
        {"newvalue": "Failure", "value": "1"},
    ],
}


class ZabbixActionTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.zabbix
    """

    def setup_loader_modules(self):
        return {zabbix_valuemap: {}}

    def test_present_create(self):
        """
        Test to ensure that named value map is created
        """
        name = "Server HP Health"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        def side_effect_run_query(*args):
            """
            Differentiate between __salt__ exec module function calls with different parameters.
            """
            if args[0] == "valuemap.get":
                return False
            elif args[0] == "valuemap.create":
                return True

        with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"valuemap": "valuemapid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=[INPUT_PARAMS, False]
                    ),
                    "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                    "zabbix.compare_params": MagicMock(return_value={}),
                },
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{0}" created.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{0}" did not exist.'.format(name),
                        "new": 'Zabbix Value map "{0}" created according definition.'.format(
                            name
                        ),
                    }
                }
                self.assertDictEqual(zabbix_valuemap.present(name, {}), ret)

    def test_present_exists(self):
        """
        Test to ensure that named value map is present and not changed
        """
        name = "Server HP Health"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"valuemap": "valuemapid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=[INPUT_PARAMS, EXISTING_OBJ]
                    ),
                    "zabbix.run_query": MagicMock(
                        return_value=["length of result is 1"]
                    ),
                    "zabbix.compare_params": MagicMock(return_value={}),
                },
            ):
                ret["result"] = True
                ret[
                    "comment"
                ] = 'Zabbix Value map "{0}" already exists and corresponds to a definition.'.format(
                    name
                )
                self.assertDictEqual(zabbix_valuemap.present(name, {}), ret)

    def test_present_update(self):
        """
        Test to ensure that named value map is present but must be updated
        """
        name = "Server HP Health"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        def side_effect_run_query(*args):
            """
            Differentiate between __salt__ exec module function calls with different parameters.
            """
            if args[0] == "valuemap.get":
                return ["length of result is 1 = valuemap exists"]
            elif args[0] == "valuemap.update":
                return DIFF_PARAMS

        with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {
                    "zabbix.get_zabbix_id_mapper": MagicMock(
                        return_value={"valuemap": "valuemapid"}
                    ),
                    "zabbix.substitute_params": MagicMock(
                        side_effect=[INPUT_PARAMS, EXISTING_OBJ_DIFF]
                    ),
                    "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                    "zabbix.compare_params": MagicMock(return_value=DIFF_PARAMS),
                },
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{0}" updated.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{0}" differed '
                        "in following parameters: {1}".format(name, DIFF_PARAMS),
                        "new": 'Zabbix Value map "{0}" fixed.'.format(name),
                    }
                }
                self.assertDictEqual(zabbix_valuemap.present(name, {}), ret)

    def test_absent_test_mode(self):
        """
        Test to ensure that named value map is absent in test mode
        """
        name = "Server HP Health"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}
        with patch.dict(zabbix_valuemap.__opts__, {"test": True}):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{0}" would be deleted.'.format(name)
                ret["changes"] = {
                    name: {
                        "old": 'Zabbix Value map "{0}" exists.'.format(name),
                        "new": 'Zabbix Value map "{0}" would be deleted.'.format(name),
                    }
                }
                self.assertDictEqual(zabbix_valuemap.absent(name), ret)

    def test_absent(self):
        """
        Test to ensure that named value map is absent
        """
        name = "Server HP Health"
        ret = {"name": name, "result": False, "comment": "", "changes": {}}
        with patch.dict(zabbix_valuemap.__opts__, {"test": False}):
            with patch.dict(
                zabbix_valuemap.__salt__,
                {"zabbix.get_object_id_by_params": MagicMock(return_value=False)},
            ):
                ret["result"] = True
                ret["comment"] = 'Zabbix Value map "{0}" does not exist.'.format(name)
                self.assertDictEqual(zabbix_valuemap.absent(name), ret)

            with patch.dict(
                zabbix_valuemap.__salt__,
                {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
            ):
                with patch.dict(
                    zabbix_valuemap.__salt__,
                    {"zabbix.run_query": MagicMock(return_value=True)},
                ):
                    ret["result"] = True
                    ret["comment"] = 'Zabbix Value map "{0}" deleted.'.format(name)
                    ret["changes"] = {
                        name: {
                            "old": 'Zabbix Value map "{0}" existed.'.format(name),
                            "new": 'Zabbix Value map "{0}" deleted.'.format(name),
                        }
                    }
                    self.assertDictEqual(zabbix_valuemap.absent(name), ret)



# -*- coding: utf-8 -*-
"""
Use Azure Blob as a Pillar source.

.. versionadded:: 3001

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure-storage-blob <https://pypi.org/project/azure-storage-blob/>`_ >= 12.0.0

The Azure Blob ext_pillar can be configured with the following parameters:

.. code-block:: yaml

    ext_pillar:
      - azureblob:
          container: 'test_container'
          connection_string: 'connection_string'
          multiple_env: False
          environment: 'base'
          blob_cache_expire: 30
          blob_sync_on_update: True

:param container: The name of the target Azure Blob Container.

:param connection_string: The connection string to use to access the specified Azure Blob Container.

:param multiple_env: Specifies whether the pillar should interpret top level folders as pillar environments.
    Defaults to false.

:param environment: Specifies which environment the container represents when in single environment mode. Defaults
    to 'base' and is ignored if multiple_env is set as True.

:param blob_cache_expire: Specifies expiration time of the Azure Blob metadata cache file. Defaults to 30s.

:param blob_sync_on_update: Specifies if the cache is synced on update. Defaults to True.

"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import pickle
import time
from copy import deepcopy

import salt.utils.files
import salt.utils.hashutils

# Import 3rd-party libs
# pylint: disable=import-error,no-name-in-module,redefined-builtin
from salt.ext import six
from salt.ext.six.moves import filter

# Import Salt libs
from salt.pillar import Pillar

# Import Azure libs
HAS_LIBS = False
try:
    from azure.storage.blob import BlobServiceClient

    HAS_LIBS = True
except ImportError:
    pass

# pylint: enable=import-error,no-name-in-module,redefined-builtin


__virtualname__ = "azureblob"

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_LIBS:
        return (
            False,
            "The following dependency is required to use the Azure Blob ext_pillar: "
            "Microsoft Azure Storage Blob >= 12.0.0 ",
        )

    return __virtualname__


def ext_pillar(
    minion_id,
    pillar,  # pylint: disable=W0613
    container,
    connection_string,
    multiple_env=False,
    environment="base",
    blob_cache_expire=30,
    blob_sync_on_update=True,
):
    """
    Execute a command and read the output as YAML.

    :param container: The name of the target Azure Blob Container.

    :param connection_string: The connection string to use to access the specified Azure Blob Container.

    :param multiple_env: Specifies whether the pillar should interpret top level folders as pillar environments.
        Defaults to false.

    :param environment: Specifies which environment the container represents when in single environment mode. Defaults
        to 'base' and is ignored if multiple_env is set as True.

    :param blob_cache_expire: Specifies expiration time of the Azure Blob metadata cache file. Defaults to 30s.

    :param blob_sync_on_update: Specifies if the cache is synced on update. Defaults to True.

    """
    # normpath is needed to remove appended '/' if root is empty string.
    pillar_dir = os.path.normpath(
        os.path.join(_get_cache_dir(), environment, container)
    )

    if __opts__["pillar_roots"].get(environment, []) == [pillar_dir]:
        return {}

    metadata = _init(
        connection_string, container, multiple_env, environment, blob_cache_expire
    )

    log.debug("Blob metadata: %s", metadata)

    if blob_sync_on_update:
        # sync the containers to the local cache
        log.info("Syncing local pillar cache from Azure Blob...")
        for saltenv, env_meta in six.iteritems(metadata):
            for container, files in six.iteritems(_find_files(env_meta)):
                for file_path in files:
                    cached_file_path = _get_cached_file_name(
                        container, saltenv, file_path
                    )
                    log.info("%s - %s : %s", container, saltenv, file_path)
                    # load the file from Azure Blob if not in the cache or too old
                    _get_file_from_blob(
                        connection_string,
                        metadata,
                        saltenv,
                        container,
                        file_path,
                        cached_file_path,
                    )

        log.info("Sync local pillar cache from Azure Blob completed.")

    opts = deepcopy(__opts__)
    opts["pillar_roots"][environment] = (
        [os.path.join(pillar_dir, environment)] if multiple_env else [pillar_dir]
    )

    # Avoid recursively re-adding this same pillar
    opts["ext_pillar"] = [x for x in opts["ext_pillar"] if "azureblob" not in x]

    pil = Pillar(opts, __grains__, minion_id, environment)

    compiled_pillar = pil.compile_pillar(ext=False)

    return compiled_pillar


def _init(connection_string, container, multiple_env, environment, blob_cache_expire):
    """
    .. versionadded:: 3001

    Connect to Blob Storage and download the metadata for each file in all containers specified and
        cache the data to disk.

    :param connection_string: The connection string to use to access the specified Azure Blob Container.

    :param container: The name of the target Azure Blob Container.

    :param multiple_env: Specifies whether the pillar should interpret top level folders as pillar environments.
        Defaults to false.

    :param environment: Specifies which environment the container represents when in single environment mode. Defaults
        to 'base' and is ignored if multiple_env is set as True.

    :param blob_cache_expire: Specifies expiration time of the Azure Blob metadata cache file. Defaults to 30s.

    """
    cache_file = _get_containers_cache_filename(container)
    exp = time.time() - blob_cache_expire

    # Check if cache_file exists and its mtime
    if os.path.isfile(cache_file):
        cache_file_mtime = os.path.getmtime(cache_file)
    else:
        # If the file does not exist then set mtime to 0 (aka epoch)
        cache_file_mtime = 0

    expired = cache_file_mtime <= exp

    log.debug(
        "Blob storage container cache file %s is %sexpired, mtime_diff=%ss, expiration=%ss",
        cache_file,
        "" if expired else "not ",
        cache_file_mtime - exp,
        blob_cache_expire,
    )

    if expired:
        pillars = _refresh_containers_cache_file(
            connection_string, container, cache_file, multiple_env, environment
        )
    else:
        pillars = _read_containers_cache_file(cache_file)

    log.debug("Blob container retrieved pillars %s", pillars)

    return pillars


def _get_cache_dir():
    """
    .. versionadded:: 3001

    Get pillar cache directory. Initialize it if it does not exist.

    """
    cache_dir = os.path.join(__opts__["cachedir"], "pillar_azureblob")

    if not os.path.isdir(cache_dir):
        log.debug("Initializing Azure Blob Pillar Cache")
        os.makedirs(cache_dir)

    return cache_dir


def _get_cached_file_name(container, saltenv, path):
    """
    .. versionadded:: 3001

    Return the cached file name for a container path file.

    :param container: The name of the target Azure Blob Container.

    :param saltenv: Specifies which environment the container represents.

    :param path: The path of the file in the container.

    """
    file_path = os.path.join(_get_cache_dir(), saltenv, container, path)

    # make sure container and saltenv directories exist
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))

    return file_path


def _get_containers_cache_filename(container):
    """
    .. versionadded:: 3001

    Return the filename of the cache for container contents. Create the path if it does not exist.

    :param container: The name of the target Azure Blob Container.

    """
    cache_dir = _get_cache_dir()
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    return os.path.join(cache_dir, "{0}-files.cache".format(container))


def _refresh_containers_cache_file(
    connection_string, container, cache_file, multiple_env=False, environment="base"
):
    """
    .. versionadded:: 3001

    Downloads the entire contents of an Azure storage container to the local filesystem.

    :param connection_string: The connection string to use to access the specified Azure Blob Container.

    :param container: The name of the target Azure Blob Container.

    :param cache_file: The path of where the file will be cached.

    :param multiple_env: Specifies whether the pillar should interpret top level folders as pillar environments.

    :param environment: Specifies which environment the container represents when in single environment mode. This is
        ignored if multiple_env is set as True.

    """
    try:
        # Create the BlobServiceClient object which will be used to create a container client
        blob_service_client = BlobServiceClient.from_connection_string(
            connection_string
        )

        # Create the ContainerClient object
        container_client = blob_service_client.get_container_client(container)
    except Exception as exc:  # pylint: disable=broad-except
        log.error("Exception: %s", exc)
        return False

    metadata = {}

    def _walk_blobs(saltenv="base", prefix=None):
        # Walk the blobs in the container with a generator
        blob_list = container_client.walk_blobs(name_starts_with=prefix)

        # Iterate over the generator
        while True:
            try:
                blob = next(blob_list)
            except StopIteration:
                break

            log.debug("Raw blob attributes: %s", blob)

            # Directories end with "/".
            if blob.name.endswith("/"):
                # Recurse into the directory
                _walk_blobs(prefix=blob.name)
                continue

            if multiple_env:
                saltenv = "base" if (not prefix or prefix == ".") else prefix[:-1]

            if saltenv not in metadata:
                metadata[saltenv] = {}

            if container not in metadata[saltenv]:
                metadata[saltenv][container] = []

            metadata[saltenv][container].append(blob)

    _walk_blobs(saltenv=environment)

    # write the metadata to disk
    if os.path.isfile(cache_file):
        os.remove(cache_file)

    log.debug("Writing Azure blobs pillar cache file")

    with salt.utils.files.fopen(cache_file, "wb") as fp_:
        pickle.dump(metadata, fp_)

    return metadata


def _read_containers_cache_file(cache_file):
    """
    .. versionadded:: 3001

    Return the contents of the containers cache file.

    :param cache_file: The path for where the file will be cached.

    """
    log.debug("Reading containers cache file")

    with salt.utils.files.fopen(cache_file, "rb") as fp_:
        data = pickle.load(fp_)

    return data


def _find_files(metadata):
    """
    .. versionadded:: 3001

    Looks for all the files in the Azure Blob container cache metadata.

    :param metadata: The metadata for the container files.

    """
    ret = {}

    for container, data in six.iteritems(metadata):
        if container not in ret:
            ret[container] = []

        # grab the paths from the metadata
        file_paths = [k["name"] for k in data]
        # filter out the dirs
        ret[container] += [k for k in file_paths if not k.endswith("/")]

    return ret


def _find_file_meta(metadata, container, saltenv, path):
    """
    .. versionadded:: 3001

    Looks for a file's metadata in the Azure Blob Container cache file.

    :param metadata: The metadata for the container files.

    :param container: The name of the target Azure Blob Container.

    :param saltenv: Specifies which environment the container represents.

    :param path: The path of the file in the container.

    """
    env_meta = metadata[saltenv] if saltenv in metadata else {}
    container_meta = env_meta[container] if container in env_meta else {}

    for item_meta in container_meta:
        item_meta = dict(item_meta)
        if "name" in item_meta and item_meta["name"] == path:
            return item_meta


def _get_file_from_blob(
    connection_string, metadata, saltenv, container, path, cached_file_path
):
    """
    .. versionadded:: 3001

    Downloads the entire contents of an Azure storage container to the local filesystem.

    :param connection_string: The connection string to use to access the specified Azure Blob Container.

    :param metadata: The metadata for the container files.

    :param saltenv: Specifies which environment the container represents when in single environment mode. This is
        ignored if multiple_env is set as True.

    :param container: The name of the target Azure Blob Container.

    :param path: The path of the file in the container.

    :param cached_file_path: The path of where the file will be cached.

    """
    # check the local cache...
    if os.path.isfile(cached_file_path):
        file_meta = _find_file_meta(metadata, container, saltenv, path)
        file_md5 = (
            "".join(list(filter(str.isalnum, file_meta["etag"]))) if file_meta else None
        )

        cached_md5 = salt.utils.hashutils.get_hash(cached_file_path, "md5")

        # hashes match we have a cache hit
        log.debug(
            "Cached file: path=%s, md5=%s, etag=%s",
            cached_file_path,
            cached_md5,
            file_md5,
        )
        if cached_md5 == file_md5:
            return

    try:
        # Create the BlobServiceClient object which will be used to create a container client
        blob_service_client = BlobServiceClient.from_connection_string(
            connection_string
        )

        # Create the ContainerClient object
        container_client = blob_service_client.get_container_client(container)

        # Create the BlobClient object
        blob_client = container_client.get_blob_client(path)
    except Exception as exc:  # pylint: disable=broad-except
        log.error("Exception: %s", exc)
        return False

    with salt.utils.files.fopen(cached_file_path, "wb") as outfile:
        outfile.write(blob_client.download_blob().readall())

    return



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import

# Import Salt Libs
import salt.states.mdadm_raid as mdadm

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MdadmTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the mdadm state
    """

    def setup_loader_modules(self):
        return {mdadm: {}}

    def test_present(self):
        """
            Test to verify that the raid is present
        """
        ret = [
            {
                "changes": {},
                "comment": "Raid salt already present.",
                "name": "salt",
                "result": True,
            },
            {
                "changes": {},
                "comment": "Devices are a mix of RAID constituents with multiple MD_UUIDs:"
                " ['6be5fc45:05802bba:1c2d6722:666f0e03', 'ffffffff:ffffffff:ffffffff:ffffffff'].",
                "name": "salt",
                "result": False,
            },
            {
                "changes": {},
                "comment": "Raid will be created with: True",
                "name": "salt",
                "result": None,
            },
            {
                "changes": {},
                "comment": "Raid salt failed to be created.",
                "name": "salt",
                "result": False,
            },
            {
                "changes": {"uuid": "6be5fc45:05802bba:1c2d6722:666f0e03"},
                "comment": "Raid salt created.",
                "name": "salt",
                "result": True,
            },
            {
                "changes": {
                    "added": ["dev1"],
                    "uuid": "6be5fc45:05802bba:1c2d6722:666f0e03",
                },
                "comment": "Raid salt assembled. Added new device dev1 to salt.\n",
                "name": "salt",
                "result": True,
            },
            {
                "changes": {"added": ["dev1"]},
                "comment": "Raid salt already present. Added new device dev1 to salt.\n",
                "name": "salt",
                "result": True,
            },
            {
                "changes": {},
                "comment": "Raid salt failed to be assembled.",
                "name": "salt",
                "result": False,
            },
        ]

        mock_raid_list_exists = MagicMock(
            return_value={"salt": {"uuid": "6be5fc45:05802bba:1c2d6722:666f0e03"}}
        )
        mock_raid_list_missing = MagicMock(return_value={})

        mock_file_access_ok = MagicMock(return_value=True)

        mock_raid_examine_ok = MagicMock(
            return_value={"MD_UUID": "6be5fc45:05802bba:1c2d6722:666f0e03"}
        )
        mock_raid_examine_missing = MagicMock(return_value={})

        mock_raid_create_success = MagicMock(return_value=True)
        mock_raid_create_fail = MagicMock(return_value=False)

        mock_raid_assemble_success = MagicMock(return_value=True)
        mock_raid_assemble_fail = MagicMock(return_value=False)

        mock_raid_add_success = MagicMock(return_value=True)

        mock_raid_save_config = MagicMock(return_value=True)

        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_exists,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_ok,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertEqual(mdadm.present("salt", 5, "dev0"), ret[0])

        mock_raid_examine_mixed = MagicMock(
            side_effect=[
                {"MD_UUID": "6be5fc45:05802bba:1c2d6722:666f0e03"},
                {"MD_UUID": "ffffffff:ffffffff:ffffffff:ffffffff"},
            ]
        )
        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_missing,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_mixed,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertEqual(mdadm.present("salt", 5, ["dev0", "dev1"]), ret[1])

        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_missing,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_missing,
                "raid.create": mock_raid_create_success,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": True}):
                self.assertDictEqual(mdadm.present("salt", 5, "dev0"), ret[2])

        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_missing,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_missing,
                "raid.create": mock_raid_create_fail,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertDictEqual(mdadm.present("salt", 5, "dev0"), ret[3])

        mock_raid_list_create = MagicMock(
            side_effect=[{}, {"salt": {"uuid": "6be5fc45:05802bba:1c2d6722:666f0e03"}}]
        )
        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_create,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_missing,
                "raid.create": mock_raid_create_success,
                "raid.save_config": mock_raid_save_config,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertDictEqual(mdadm.present("salt", 5, "dev0"), ret[4])

        mock_raid_examine_replaced = MagicMock(
            side_effect=[{"MD_UUID": "6be5fc45:05802bba:1c2d6722:666f0e03"}, {}]
        )
        mock_raid_list_create = MagicMock(
            side_effect=[{}, {"salt": {"uuid": "6be5fc45:05802bba:1c2d6722:666f0e03"}}]
        )
        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_create,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_replaced,
                "raid.assemble": mock_raid_assemble_success,
                "raid.add": mock_raid_add_success,
                "raid.save_config": mock_raid_save_config,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertDictEqual(mdadm.present("salt", 5, ["dev0", "dev1"]), ret[5])

        mock_raid_examine_replaced = MagicMock(
            side_effect=[{"MD_UUID": "6be5fc45:05802bba:1c2d6722:666f0e03"}, {}]
        )
        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_exists,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_replaced,
                "raid.add": mock_raid_add_success,
                "raid.save_config": mock_raid_save_config,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertDictEqual(mdadm.present("salt", 5, ["dev0", "dev1"]), ret[6])

        mock_raid_examine_replaced = MagicMock(
            side_effect=[{"MD_UUID": "6be5fc45:05802bba:1c2d6722:666f0e03"}, {}]
        )
        with patch.dict(
            mdadm.__salt__,
            {
                "raid.list": mock_raid_list_missing,
                "file.access": mock_file_access_ok,
                "raid.examine": mock_raid_examine_replaced,
                "raid.assemble": mock_raid_assemble_fail,
            },
        ):
            with patch.dict(mdadm.__opts__, {"test": False}):
                self.assertDictEqual(mdadm.present("salt", 5, ["dev0", "dev1"]), ret[7])

    def test_absent(self):
        """
            Test to verify that the raid is absent
        """
        ret = [
            {
                "changes": {},
                "comment": "Raid salt already absent",
                "name": "salt",
                "result": True,
            },
            {
                "changes": {},
                "comment": "Raid saltstack is set to be destroyed",
                "name": "saltstack",
                "result": None,
            },
            {
                "changes": {},
                "comment": "Raid saltstack has been destroyed",
                "name": "saltstack",
                "result": True,
            },
        ]

        mock = MagicMock(return_value=["saltstack"])
        with patch.dict(mdadm.__salt__, {"raid.list": mock}):
            self.assertDictEqual(mdadm.absent("salt"), ret[0])

            with patch.dict(mdadm.__opts__, {"test": True}):
                self.assertDictEqual(mdadm.absent("saltstack"), ret[1])

            with patch.dict(mdadm.__opts__, {"test": False}):
                mock = MagicMock(return_value=True)
                with patch.dict(mdadm.__salt__, {"raid.destroy": mock}):
                    self.assertDictEqual(mdadm.absent("saltstack"), ret[2])
